#!/usr/bin/env python3
# -*- coding: utf-8 -*

import abc
import copy
from typing import Optional, Union, Tuple, List, NoReturn

import numpy as np

from ase.calculators.calculator import Calculator

from .. import config
from ..core.register import registers
from ..computation import register_drivers
from .calculators.dummy import DummyCalculator


"""The abstract base class of any potential manager.

"""


class AbstractPotentialManager(abc.ABC):
    """
    Create various potential instances
    """

    #: Name of the potential.
    name: str = "potential"

    #: Supported calculator backends.
    implemented_backends: List[str] = []

    #: Supported combinations of calculator backend and driver/engine.
    valid_combinations: tuple = ()

    def __init__(self):
        """ """
        #: Attached calculator.
        self._calc: Calculator = DummyCalculator()

        return

    @property
    def calc(self) -> Calculator:
        """Attached calculator."""
        return self._calc

    @calc.setter
    def calc(self, calc_):
        self._calc = calc_
        return

    @property
    def calc_backend(self) -> str:
        """Backend of attached calculator."""

        return self.calc_params.get("backend", self.name)

    @abc.abstractmethod
    def register_calculator(self, calc_params: dict, *agrs, **kwargs):
        """Register the host calculator."""
        # Save the original copy of calc_params and pop the backend keyword
        # as it is not for calculator.
        self.calc_params = copy.deepcopy(calc_params)
        calc_params.pop("backend")

        if self.calc_backend not in self.implemented_backends:
            raise RuntimeError(
                f"Unknown backend {self.calc_backend} for potential {self.name} with {self.implemented_backends}."
            )

        return

    def create_driver(self, dyn_params: dict = {}, *args, **kwargs):
        """Create a driver for dynamics.

        Default the dynamics backend will be the same as calc. However,
        ase-based dynamics can be used for all calculators.

        """
        # - check whether there is a calc
        if not hasattr(self, "calc"):
            raise AttributeError(
                "Cannot create driver since a calculator has been properly registered."
            )

        # parse backends
        self.dyn_params = dyn_params
        dynamics = dyn_params.get("backend", self.calc_backend)
        if dynamics == "external":
            dynamics = self.calc_backend

        if (self.calc_backend, dynamics) not in self.valid_combinations:
            raise RuntimeError(
                f"Invalid dynamics backend {dynamics} based on {self.calc_backend} calculator"
            )

        # - merge params for compat
        merged_params = {}
        if "task" in dyn_params:
            merged_params.update(task=dyn_params.get("task", "min"))

        if "init" in dyn_params or "run" in dyn_params:
            merged_params.update(**dyn_params.get("init", {}))
            merged_params.update(**dyn_params.get("run", {}))
        else:
            merged_params.update(**dyn_params)

        # -- add params from key besides task, init, and run
        merged_params.update(
            ignore_convergence=dyn_params.get("ignore_convergence", False),
            random_seed=dyn_params.get("random_seed", None),
        )

        # -- other params
        ignore_convergence = merged_params.pop("ignore_convergence", False)

        # TODO: make PotentialManager a Node as well???
        random_seed = merged_params.pop(
            "random_seed", int(config.GRNG.integers(0, 1e8))
        )

        # - create dynamics
        driver_cls = register_drivers[dynamics]
        # assert driver_cls is not None, f"Cannot find a driver named {dynamics}."

        driver = driver_cls(
            self.calc,
            merged_params,
            directory=self.calc.directory,
            ignore_convergence=ignore_convergence,
            random_seed=random_seed,
        )
        driver.pot_params = self.as_dict()

        return driver

    def create_reactor(self, rxn_params: dict = {}, *args, **kwargs):
        """Create a reactor for reaction.

        Default the reaction backend will be the same as calc. However,
        ase-based dynamics can be used for all calculators.

        """
        # - check whether there is a calc
        if not hasattr(self, "calc"):
            raise AttributeError(
                "Cant create reactor before a calculator has been properly registered."
            )

        # parse backends
        self.rxn_params = rxn_params
        reaction = rxn_params.get("backend", self.calc_backend)
        if reaction == "external":
            reaction = self.calc_backend

        if (self.calc_backend, reaction) not in self.valid_combinations:
            raise RuntimeError(
                f"Invalid reaction backend {reaction} based on {self.calc_backend} calculator. Valid combinations are {self.valid_combinations}"
            )

        # - merge params for compat
        merged_params = {}
        if "task" in rxn_params:
            merged_params.update(task=rxn_params.get("task", "min"))
        if "init" in rxn_params or "run" in rxn_params:
            merged_params.update(**rxn_params.get("init", {}))
            merged_params.update(**rxn_params.get("run", {}))
        else:
            merged_params.update(**rxn_params)

        # - other params
        ignore_convergence = merged_params.pop("ignore_convergence", False)

        # - construct driver params
        inp_params = dict(
            calc=self.calc, params=merged_params, ignore_convergence=ignore_convergence
        )

        driver = registers.create("reactor", reaction, convert_name=False, **inp_params)
        driver.pot_params = self.as_dict()

        return driver

    def as_dict(self):
        """"""
        params = {}
        params["name"] = self.name
        params["params"] = copy.deepcopy(self.calc_params)

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from ..core.variable import Variable, DummyVariable
from ..core.operation import Operation
from ..core.register import registers

from ..potential.managers.mixer import MixerManager


@registers.variable.register
class PotterVariable(Variable):

    def __init__(self, directory="./", **kwargs):
        """"""
        # manager = PotentialRegister()
        name = kwargs.get("name", None)
        # potter = manager.create_potential(pot_name=name)
        # potter.register_calculator(kwargs.get("params", {}))
        # potter.version = kwargs.get("version", "unknown")

        potter = registers.create(
            "manager",
            name,
            convert_name=True,
            # **kwargs.get("params", {})
        )
        potter.register_calculator(kwargs.get("params", {}))

        super().__init__(initial_value=potter, directory=directory)

        return


def create_mixer(basic_params, *args, **kwargs):
    """"""
    potters = [basic_params]
    for x in args:
        potters.append(x)
    calc_params = dict(backend="ase", potters=potters)

    mixer = MixerManager()
    mixer.register_calculator(calc_params=calc_params)

    return mixer


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ase.ga.standardmutations import RattleMutation, PermutationMutation, MirrorMutation
from ase.ga.particle_mutations import (
    RandomMutation, RandomPermutation, COM2surfPermutation, Poor2richPermutation, 
    Rich2poorPermutation, SymmetricSubstitute, RandomSubstitute
)
from ase.ga.standardmutations import StrainMutation
from ase.ga.soft_mutation import SoftMutation, BondElectroNegativityModel

from gdpx.core.register import registers

# - standard
registers.builder.register("rattle")(RattleMutation)
registers.builder.register("permutation")(PermutationMutation)
registers.builder.register("mirror")(MirrorMutation)

# - bulk
registers.builder.register("strain")(StrainMutation)
registers.builder.register("soft")(SoftMutation)

# - cluster
#registers.builder.register("random")(RandomMutation)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import pathlib
from typing import Union

import numpy as np

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from gdpx.utils.comparasion import parity_plot, parity_plot_dict, PropInfo


import matplotlib as mpl
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except:
    ...


def use_dpeval():
    import matplotlib as mpl
    mpl.use('Agg') #silent mode
    from matplotlib import pyplot as plt
    plt.style.use('presentation')

    import deepmd.DeepPot as DeepPot

    type_map = {"O": 0, "Pt": 1}
    inverse_type_map = {0: "O", 1: "Pt"}

    raw_dirs = pathlib.Path('../oxides/gdp-main/it-0011/raw_data')

    # find system
    sys_map = {"O": 1, "Pt": 36}
    sys_name = []
    for k, v in sys_map.items():
        sys_name.extend([str(k), str(v)])
    sys_name = ''.join(sys_name)

    test = raw_dirs / sys_name
    print(test)

    natoms = np.sum(list(sys_map.values())) 

    set_dirs = list(test.glob('set*'))
    set_dirs.sort()

    # train data
    train_boxes, train_coords, train_energies, train_forces = [], [], [], []
    for set_dir in set_dirs[:-1]:
        train_boxes.extend( np.load(set_dir / 'box.npy').tolist() )
        train_coords.extend( np.load(set_dir / 'coord.npy').tolist() )
        train_energies.extend( np.load(set_dir / 'energy.npy').tolist() )
        train_forces.extend( np.load(set_dir / 'force.npy').tolist() )
        #print(boxes.shape)
        #print(boxes)

    # test_data
    test_boxes, test_coords, test_energies, test_forces = [], [], [], []
    for set_dir in set_dirs[-1:]:
        test_boxes.extend( np.load(set_dir / 'box.npy').tolist() )
        test_coords.extend( np.load(set_dir / 'coord.npy').tolist() )
        test_energies.extend( np.load(set_dir / 'energy.npy').tolist() )
        test_forces.extend( np.load(set_dir / 'force.npy').tolist() )

    # read types
    atype = np.loadtxt(test / 'type.raw', dtype=int)
    print(atype)

    # start with a square Figure
    fig, axarr = plt.subplots(nrows=3, ncols=2, figsize=(16,16))
    axarr = axarr.flatten()

    plt.suptitle('Dataset Overview')
    #plt.tight_layout()
    plt.subplots_adjust(
        left=0.10, right=0.95,
        bottom=0.10, top=0.85,
        wspace=0.20, hspace=0.30
    )

    plot_hist(axarr[0], np.array(train_energies).flatten()/natoms, 'Energy [eV]', 'Number of Frames')
    plot_hist(axarr[1], np.array(train_forces).flatten(), 'Force [eV/AA]', 'Number of Frames')
    plot_hist(axarr[0], np.array(test_energies).flatten()/natoms, 'Energy [eV]', 'Number of Frames')
    plot_hist(axarr[1], np.array(test_forces).flatten(), 'Force [eV/AA]', 'Number of Frames')

    # use dp to check test
    dp = DeepPot('../oxides/gdp-main/it-0011/ensemble/model-0/graph.pb')
    dp_input = {'coords': test_coords, 'cells': test_boxes, 'atom_types': atype, 'atomic': True}
    e, f, v, ae, av = dp.eval(**dp_input)

    prop_info = PropInfo(xlabel='miaow', ylabel='miaow', title='xxx')
    parity_plot(['DP', e.flatten()/natoms], ['DFT', np.array(test_energies).flatten()/natoms], axarr[2], ('energy', '[eV/atom]'), sys_name)
    #parity_plot(['DP', f.flatten()], ['DFT', np.array(test_forces).flatten()], axarr[3], ('force', '[eV/AA]'), sys_name)
    parity_plot_dict(transform_forces(atype, f.reshape(-1,natoms*3)), transform_forces(atype, test_forces), axarr[3], prop_info)

    dp_input = {'coords': train_coords, 'cells': train_boxes, 'atom_types': atype, 'atomic': True}
    e, f, v, ae, av = dp.eval(**dp_input)

    #parity_plot(['DP', e.flatten()/natoms], ['DFT', np.array(test_energies).flatten()/natoms], axarr[4], ('energy', '[eV/atom]'), sys_name)
    #parity_plot_dict(transform_forces(atype, f.reshape(-1,natoms*3)), transform_forces(atype, test_forces), axarr[5], prop_info)

    plt.savefig('wang.png')

    return

def set2inputs(set_dir):
    """ extract info from dpdata
    """
    boxes = np.load(set_dir / 'box.npy')
    coords = np.load(set_dir / 'coord.npy')
    energies = np.load(set_dir / 'energy.npy')
    forces = np.load(set_dir / 'force.npy')

    nframes = boxes.shape[0]
    boxes = boxes.reshape(-1,3,3)
    #boxes = torch.from_numpy(boxes)
    positions = coords.reshape(nframes,-1,3)
    #positions = torch.from_numpy(positions)
    natoms = positions.shape[1]
    energies = energies.flatten().tolist()
    forces = forces.flatten().tolist()

    return (
        nframes, natoms, 
        energies, forces, # 1D-List
        positions, boxes # tensor
    )

def sets2results(set_dirs, calc, chemical_symbols):
    """ test set_dirs
    """
    ref_energies, ref_forces = [], []
    mlp_energies, mlp_forces = [], []
    for set_dir in set_dirs:
        set_start_time = time.time()
        (
            nframes, natoms, 
            energies, forces,
            positions, boxes
        ) = set2inputs(set_dir)
        ref_energies.extend(energies)
        ref_forces.extend(forces)

        print('nframes in train', nframes)
        for i in range(nframes):
            atoms = Atoms(
                symbols=chemical_symbols, positions=positions[i], 
                cell=boxes[i], pbc=[1,1,1]
            )
            #results = {'energy': energies[i], 'forces': forces[i,:].reshape(-1,3)}
            #spc = SinglePointCalculator(atoms, **results)
            #atoms.calc = spc
            calc.reset()
            atoms.calc = calc
            #print(type(atoms.get_potential_energy()))
            #print(type(atoms.get_forces()))
            mlp_energies.append(float(atoms.get_potential_energy()))
            mlp_forces.extend(atoms.get_forces().flatten().tolist())
        set_end_time = time.time()
        print('test time: ', set_end_time - set_start_time)

    return ref_energies, ref_forces, mlp_energies, mlp_forces

def append_predictions(
    frames, calc = None, other_props = [], calc_uncertainty=False
):
    """ test xyz
    """
    if calc is not None:
        calc_name = calc.name.lower()

    tot_energies, tot_forces = [], {}
    tot_props = {}

    for atoms in frames: # free energy per atom
        # set calculator
        calc_atoms = atoms.copy()
        if calc is not None:
            calc.reset()
            if calc_uncertainty:
                calc.calc_uncertainty = True # EANN specific
            calc_atoms.calc = calc
            new_forces = calc_atoms.get_forces(apply_constraint=False)
            new_energy = calc_atoms.get_potential_energy()
            atoms.info[calc_name+"_energy"] = new_energy
            atoms.arrays[calc_name+"_forces"] = new_forces.copy()
        
    return frames

def merge_forces(frames):
    """ convert forces of all frames into a dict,
        which helps further comparison
    """
    tot_forces = {}
    for atoms in frames: # free energy per atom
        # basic info
        symbols = atoms.get_chemical_symbols()

        # force
        forces = atoms.get_forces().copy()
        for sym, force in zip(symbols, forces.tolist()):
            if sym in tot_forces.keys():
                tot_forces[sym].extend(force)
            else:
                tot_forces[sym] = force

    return tot_forces

def merge_predicted_forces(frames, calc_name):
    """ convert forces of all frames into a dict,
        which helps further comparison
    """
    tot_forces = {}
    for atoms in frames: # free energy per atom
        # basic info
        symbols = atoms.get_chemical_symbols()

        # force
        forces = atoms.arrays[calc_name+"_forces"].copy()
        for sym, force in zip(symbols, forces.tolist()):
            if sym in tot_forces.keys():
                tot_forces[sym].extend(force)
            else:
                tot_forces[sym] = force

    return tot_forces


def xyz2results(frames, calc = None, other_props = []):
    """ test xyz
    """
    #new_frames = frames.copy()

    tot_energies, tot_forces = [], {}
    tot_props = {}

    for atoms in frames: # free energy per atom
        # basic info
        symbols = atoms.get_chemical_symbols()

        # set calculator
        if calc is not None:
            calc.reset()
            calc.calc_uncertainty = True # EANN specific
            atoms.calc = calc

        # energy
        energy = atoms.get_potential_energy() 
        tot_energies.append(energy)

        # force
        forces = atoms.get_forces(apply_constraint=False)
        for sym, force in zip(symbols, forces.tolist()):
            if sym in tot_forces.keys():
                tot_forces[sym].extend(force)
            else:
                tot_forces[sym] = force
        
        # other properties
        if calc is not None:
            for prop in other_props:
                if prop in tot_props.keys():
                    tot_props[prop].extend([atoms.calc.results[prop]])
                else:
                    tot_props[prop] = [atoms.calc.results[prop]]

    if tot_props:
        return tot_energies, tot_forces, tot_props
    else:
        return tot_energies, tot_forces

def calc_and_compare_results(frames, calc):
    """"""
    ref_energies, ref_forces = xyz2results(frames, calc=None)
    natoms_array = np.array([len(x) for x in frames])

    calc_name = calc.name.lower()
    #print("Calculating with MLP...")
    new_frames = append_predictions(frames, calc=calc)
    #mlp_energies, mlp_forces = xyz2results(frames, calc=calc)
    mlp_energies = [a.info[calc_name+"_energy"] for a in new_frames]
    mlp_forces = merge_predicted_forces(new_frames, calc_name) 

    forces = np.array([ref_forces, mlp_forces])
    energies = np.array([ref_energies, mlp_energies]) / natoms_array

    return energies, forces


def plot_comparasion(calc_name, energies, forces, saved_figure):
    """"""
    fig, axarr = plt.subplots(
        nrows=1, ncols=2,
        gridspec_kw={'hspace': 0.3}, figsize=(16, 12)
    )
    axarr = axarr.flatten()

    _, nframes = energies.shape
    plt.suptitle(f"Number of frames {nframes}")

    ax = axarr[0]
    en_rmse_results = parity_plot_dict(
        {"energy": energies[0, :]}, {"energy": energies[1, :]}, 
        ax, 
        {
            "xlabel": "DFT [eV]", "ylabel": "%s [eV]" % calc_name, "title": "Energy"
        }
    )

    ax = axarr[1]
    for_rmse_results = parity_plot_dict(
        forces[0], forces[1],
        ax,
        {
            "xlabel": "DFT [eV/AA]",
            "ylabel": "%s [eV/AA]" % calc_name,
            "title": "Forces"
        }
    )

    plt.savefig(saved_figure)

    return en_rmse_results, for_rmse_results


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
See https://demonstrations.wolfram.com/TrajectoriesOnTheMullerBrownPotentialEnergySurface/#more
K. Müller and L. D. Brown, 
"Location of Saddle Points and Minimum Energy Paths by a Constrained Simplex Optimization Procedure," 
Theoretical Chemistry Accounts, 53, 1979 pp. 75–93.

Wolfe-Quapp PES
"""

from termios import VEOL
import torch
import numpy as np

def pes_np(x, y):
    # parameters
    A =  [-200, -100, -170, 15]
    a =  [-1, -1, -6.5, 0.7]
    b =  [0, 0, 11, 0.6]
    c =  [-10, -10, -6.5, 0.7]
    x0 = [1, 0, -0.5, -1]
    y0 = [0, 0.5, 1.5, 1]

    energy = 0.0
    for k in range(4):
        energy += (
            A[k] * np.exp(a[k]*(x-x0[k])**2 + 
            b[k]*(x-x0[k])*(y-y0[k]) + 
            c[k]*(y-y0[k])**2)
        )

    return energy

def MullerBrownPES(pos):
    # unpack position and require gradients
    pos.requires_grad = True

    # parameters
    A =  torch.tensor([-200, -100, -170, 15])
    a =  torch.tensor([-1, -1, -6.5, 0.7])
    b =  torch.tensor([0, 0, 11, 0.6])
    c =  torch.tensor([-10, -10, -6.5, 0.7])
    x0 = torch.tensor([1, 0, -0.5, -1])
    y0 = torch.tensor([0, 0.5, 1.5, 1])

    x = pos[:, 0].reshape(-1,1)
    y = pos[:, 1].reshape(-1,1)

    dx = x - x0
    dy = y - y0

    energy = torch.sum(
        A * torch.exp(a*dx**2 + b*dx*dy + c*dy**2), 
        dim=-1
    )
    energy_list = [e for e in energy]
    forces = -torch.autograd.grad(energy_list, pos, retain_graph=False)[0]

    return energy, forces

if __name__ == '__main__':
    #pos = torch.tensor([0.5, 0.5], requires_grad=True)
    pos = torch.tensor(
        [
            [0.5, 0.5],
            [1.5, 1.5]
        ], 
        requires_grad=False
    )
    print(pos)
    energy, forces = MullerBrownPES(pos)
    print(energy)
    print(forces)

    # plot data
    import matplotlib as mpl
    mpl.use('Agg')  # silent mode
    from matplotlib import pyplot as plt
    plt.style.use('presentation')

    import matplotlib.cm as cm
    delta = 0.025
    x = np.arange(-1.6, 1.6, delta)
    y = np.arange(-0.8, 2.4, delta)
    X, Y = np.meshgrid(x, y)
    print(X.shape)
    print(Y.shape)

    positions = np.array([[xs, ys] for xs, ys in zip(X.flatten(),Y.flatten())])
    positions = torch.from_numpy(positions)
    Z, _ = MullerBrownPES(positions)
    Z = Z.detach().numpy()
    Z = Z.reshape(X.shape)

    fig, ax = plt.subplots(figsize=(16,12), constrained_layout=False)

    plt.suptitle("Muller-Brown PES")

    levels = np.arange(-160, 240, 40).tolist()
    levels.extend([320,400])
    cset1 = ax.contourf(X, Y, Z, levels)
    cset2 = plt.contour(X, Y, Z, cset1.levels, colors='k')

    ax.clabel(cset2, inline=True, fmt="%d")

    # Make a colorbar for the ContourSet returned by the contourf call.
    cbar = fig.colorbar(cset1)
    cbar.ax.set_ylabel("Potential Energy")
    # Add the contour line levels to the colorbar
    # cbar.add_lines(cset1)

    plt.tight_layout()
    plt.savefig("pes.png")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import networkx as nx

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
#plt.style.use("presentation")

def node_symbol(symbol, idx, offset):
    """"""
    return "{}:{}:[{},{},{}]".format(symbol, idx, offset[0], offset[1], offset[2])

def unpack_node_name(node_symbol):
    #print("xxx: ", node_symbol)
    chem_sym, idx, offset = node_symbol.split(":")
    idx = int(idx)
    offset = eval(offset)

    return chem_sym, idx, offset

def bond_symbol(sym1, sym2, a1, a2):
    return "{}{}".format(*sorted((sym1, sym2)))

def grid_iterator(grid):
    """Yield all of the coordinates in a 3D grid as tuples

    Args:
        grid (tuple[int] or int): The grid dimension(s) to
                                  iterate over (x or (x, y, z))

    Yields:
        tuple: (x, y, z) coordinates
    """
    if isinstance(grid, int): # Expand to 3D grid
        grid = (grid, grid, grid)

    for x in range(-grid[0], grid[0]+1):
        for y in range(-grid[1], grid[1]+1):
            for z in range(-grid[2], grid[2]+1):
                yield (x, y, z)

def show_edges(graph):
    print("----- See Edges -----")
    for (u, v, d) in graph.edges.data():
        print(u, v, d)
    
    return
        
def show_nodes(graph):
    print("----- See Nodes -----")
    for (u, d) in graph.nodes.data():
        print(u, d)
    
    return

def show_components():
    #print("----- connected components -----")
    #for c in nx.connected_components(graph):
    #    print(c)
    pass

    return

def plot_graph(graph, fig_name="graph.png"):
    # plot graph
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
    ax.set_title("Graph")

    nx.draw(graph, with_labels=True)

    plt.savefig(fig_name)

    return

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
import warnings
import pathlib
import re
from typing import List, Union

import numpy as np
import numpy.ma as ma

from ase import Atoms
from ase.io import read, write

import matplotlib.pyplot as plt

try:
    plt.style.use("presentation")
except Exception as e:
    ...


from ..worker.drive import DriverBasedWorker
from ..utils.comparision import get_properties, plot_parity, plot_distribution
from ..utils.command import convert_indices

from .validator import AbstractValidator


class SinglepointValidator(AbstractValidator):
    """Calculate energies on each structures and save them to file."""

    def __init__(self, groups: dict = None, convergence: dict = None, *args, **kwargs):
        """Init a spc validator.

        Args:
            groups: Errors are estimated for given groups.

        """
        super().__init__(*args, **kwargs)
        self.groups = groups
        self.convergence = convergence

        return

    def run(self, dataset, worker: DriverBasedWorker, *args, **kwargs):
        """"""
        super().run()

        data = []
        frame_pairs = []
        for prefix, frames in dataset["reference"]:
            pred_frames = self._irun(prefix, frames, None, worker)
            nframes, rmse_ret = self._plot_comparison(prefix, frames, pred_frames)
            frame_pairs.append([frames, pred_frames])
            data.append([prefix, nframes, rmse_ret])
        self.write_data(data)

        # - plot specific groups
        group_params = copy.deepcopy(self.groups)

        def run_selection():
            selected_prefixes, selected_groups = [], []
            for k, v in group_params.items():
                selected_prefixes.append(k)
                selected_groups.append(convert_indices(v, index_convention="lmp"))
            self._debug(selected_groups)
            self._debug(selected_prefixes)

            for curr_prefix, curr_indices in zip(selected_prefixes, selected_groups):
                curr_ref = list(
                    itertools.chain(*[frame_pairs[i][0] for i in curr_indices])
                )
                curr_pre = list(
                    itertools.chain(*[frame_pairs[i][1] for i in curr_indices])
                )
                nframes, rmse_ret = self._plot_comparison(
                    curr_prefix, curr_ref, curr_pre
                )
                self.write_data(
                    [[curr_prefix, nframes, rmse_ret]], f"{curr_prefix}-rmse.dat"
                )

        if group_params is not None:
            run_selection()

        return

    def write_data(self, data, fname: str = "rmse.dat"):
        """"""
        # - check data file
        keys = ["ene", "frc"]
        for rmse_ret in [x[2] for x in data]:
            for k in rmse_ret.keys():
                if k not in keys:
                    keys.append(k)
        content_fmt = "{:<48s}  {:>8d}  " + "{:>8.4f}  {:>8.4f}  " * len(keys) + "\n"

        header_fmt = "{:<48s}  {:>8s}  " + "{:>8s}  {:>8s}  " * len(keys) + "\n"
        header_data = ["#prefix", "nframes"]
        for k in keys:
            header_data.extend([f"{k}_rmse", f"{k}_std"])
        header = header_fmt.format(*header_data)

        content = header
        for prefix, nframes, rmse_ret in data:
            cur_data = [prefix, nframes]
            for k in keys:
                v = rmse_ret.get(k, None)
                if v is None:
                    cur_data.extend([np.nan, np.nan])
                else:
                    cur_data.extend([v["rmse"], v["std"]])
            content += content_fmt.format(*cur_data)

        with open(self.directory / fname, "w") as fopen:
            fopen.write(content)
        for l in content.split("\n"):
            self._print(l)

        return

    def _irun(
        self, prefix: str, ref_frames: List[Atoms], pred_frames: List[Atoms], worker
    ):
        """"""
        # - read structures
        nframes = len(ref_frames)
        if pred_frames is None:
            # NOTE: use worker to calculate
            # TODO: use cached data?
            self._print(f"Calculate reference frames {prefix} with potential...")
            cached_pred_fpath = self.directory / prefix / "pred.xyz"
            if not cached_pred_fpath.exists():
                worker.directory = self.directory / prefix
                worker.batchsize = nframes

                worker._share_wdir = True

                worker.run(ref_frames)
                worker.inspect(resubmit=True)
                if worker.get_number_of_running_jobs() == 0:
                    ret = worker.retrieve(
                        include_retrieved=True,
                    )
                    pred_frames = list(itertools.chain(*ret))
                else:
                    # TODO: ...
                    ...
                write(cached_pred_fpath, pred_frames)
            else:
                pred_frames = read(cached_pred_fpath, ":")
        else:
            ...

        return pred_frames

    def _plot_comparison(
        self, prefix, ref_frames: List[Atoms], pred_frames: List[Atoms]
    ):
        """"""
        if not (self.directory / prefix).exists():
            (self.directory / prefix).mkdir(parents=True)

        nframes = len(ref_frames)
        ref_symbols, ref_energies, ref_forces = get_properties(ref_frames)
        ref_natoms = [len(a) for a in ref_frames]
        pred_symbols, pred_energies, pred_forces = get_properties(pred_frames)

        # - figure
        fig, axarr = plt.subplots(
            nrows=1, ncols=2, gridspec_kw={"hspace": 0.3}, figsize=(16, 9)
        )
        axarr = axarr.flatten()
        plt.suptitle(f"{prefix} with nframes {nframes}")

        # -- energies
        ene_rmse = plot_parity(
            axarr[0], ref_energies, pred_energies, x_name="ene", weights=ref_natoms
        )

        # -- forces
        frc_rmse = plot_parity(
            axarr[1], ref_forces, pred_forces, x_name="frc", x_types=ref_symbols
        )

        # if (self.directory/f"{prefix}.png").exists():
        #    warnings.warn(f"Figure file {prefix} exists.", UserWarning)
        plt.savefig(self.directory / prefix / "rmse.png")
        plt.close()

        # plot distributions
        fig, axarr = plt.subplots(
            nrows=1, ncols=2, gridspec_kw={"hspace": 0.3}, figsize=(16, 9)
        )
        axarr = axarr.flatten()
        plt.suptitle(f"{prefix} with nframes {nframes}")

        plot_distribution(
            axarr[0], ref_energies, pred_energies, x_name="ene", weights=ref_natoms
        )
        plot_distribution(
            axarr[1], ref_forces, pred_forces, x_name="frc", x_types=ref_symbols
        )

        plt.savefig(self.directory / prefix / "dist.png")
        plt.close()

        # - save results to data file
        rmse_ret = {}
        x_rmse, x_rmse_names = ene_rmse
        for _rms, rms_name in zip(x_rmse, x_rmse_names):
            rmse_ret[rms_name] = _rms
        x_rmse, x_rmse_names = frc_rmse
        for _rms, rms_name in zip(x_rmse, x_rmse_names):
            rmse_ret[rms_name] = _rms

        return nframes, rmse_ret

    def report_convergence(self, *args, **kwargs):
        """"""
        converged = True

        rmse_fpath = self.directory / "rmse.dat"
        if rmse_fpath.exists():
            with open(rmse_fpath, "r") as fopen:
                lines = fopen.readlines()
            col_names = lines[0].strip()[1:].split()[1:]
            row_names = [x.strip().split()[0] for x in lines[1:]]
            data = np.array(
                [x.strip().split()[1:] for x in lines[1:]], dtype=np.float32
            )

            # --
            convergence = copy.deepcopy(self.convergence)
            pattern = convergence.pop("pattern", None)
            if pattern is not None:
                matched_names = [x for x in row_names if re.match(pattern, x)]
            else:
                matched_names = row_names

            assert all(
                [x in col_names for x in convergence.keys()]
            ), "Unavailable keys for convergence."

            converged = True
            for name in matched_names:
                self._print(name)
                iname = row_names.index(name)
                for k, v in convergence.items():
                    ik = col_names.index(k)
                    self._print(f"{data[iname, ik]} <=? {v}")
                    if data[iname, ik] <= v:
                        converged = True
                    else:
                        converged = False
        else:
            self._print("No rmse data is available and set convergence to True.")

        self._print(f"    >>> {converged}")

        return converged


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List

import numpy as np 

from ase import Atoms
from ase.io import read, write 

from .builder import StructureModifier

"""Cleave surface from the bulk structure.
Surface miller index <= 2 
"""


class CleaveSurfaceModifier(StructureModifier):

    name: str = "cleave_surface"

    def __init__(
            self, backend="ase", max_miller: int=2, slab_size: float=4.0, vaccum_size: float=12.0, 
            center_slab=False, normal=1, substrates=None, *args, **kwargs
        ):
        """"""
        super().__init__(substrates, *args, **kwargs)

        self.backend = backend

        self.max_miller = max_miller
        self.slab_size = slab_size
        self.vaccum_size = vaccum_size
        self.center_slab = center_slab
        self.normal = normal

        return
    
    def run(self, substrates=None, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        frames = []
        if self.backend == "ase":
            from ase.lattice.cubic import FaceCenteredCubic
            #atoms = FaceCenteredCubic(
            #    symbol="Cu", 
            #    #directions=[[1,-1,0], [1,1,-2], [1,1,1]],
            #    directions=[[1,0,0], [1,1,-2], [1,1,1]],
            #    debug=2
            #)
            raise NotImplementedError()
        elif self.backend == "matgen":
            try:
                import pymatgen as mg
                import pymatgen.core.surface as mg_surface 
                from pymatgen.io.ase import AseAtomsAdaptor
                from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
            except Exception as e:
                raise RuntimeError(e)
            for substrate in self.substrates:
                structure = AseAtomsAdaptor.get_structure(substrate)
                structure = SpacegroupAnalyzer(structure).get_conventional_standard_structure()
                print(structure.lattice.d_hkl((1,1,1)))
                possible_slabs = mg_surface.generate_all_slabs(
                    structure, 
                    max_index = self.max_miller, 
                    min_slab_size = self.slab_size, 
                    min_vacuum_size = self.vaccum_size, 
                    center_slab = self.center_slab,
                    max_normal_search = self.normal,
                    in_unit_planes=True
                )
                frames.extend([AseAtomsAdaptor.get_atoms(s) for s in possible_slabs])
        else:
            ...

        return frames
    
    def _irun(self):
        """"""

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""This module aims to offer a unified interface to the computation of descriptors."""

from ..core.register import registers

# ---
from .interface import DescriberVariable, describe
registers.variable.register(DescriberVariable)
registers.operation.register(describe)

# ---
from .spc import SpcDescriber
registers.describer.register("spc")(SpcDescriber)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import Union, List

import numpy as np
from joblib import Parallel, delayed

from ase import Atoms

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ..utils.command import CustomTimer
from .validator import AbstractValidator
from .utils import wrap_traj
from ..builder.group import create_a_group
from ..data.array import AtomsNDArray


def plot_msd(wdir, names, lagtimes, timeseries, start_step=20, end_step=60, prefix=""):
    """"""
    # - get self-diffusivity
    from scipy.stats import linregress

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
    fig.suptitle("MSD")

    #ax.set_title("$51Å\times 44Å$")
    if names is None:
        names = [str(i) for i in range(len(lagtimes))]

    for name, x, y in zip(names, lagtimes, timeseries):

        linear_model = linregress(
            x[start_step:end_step], y[start_step:end_step]
        )

        slope = linear_model.slope
        error = linear_model.rvalue
        # dim_fac is 3 as we computed a 3D msd with 'xyz'
        D = slope * 1/(2*3)
        print(f"{name} D: ", D)

        #ax.plot(x, y, label=f"{name} K $D={D:>.2e}$")
        ax.plot(x, y, label=f"{name}")

        ax.text(np.median(x), np.median(y), f"$D={D:>.2e}$")


    ax.set_ylabel("MSD [Å^2]")
    ax.set_xlabel("Time [ps]")

    ax.legend()

    plt.savefig(wdir/ f"{prefix}msd.png")

    return


def compute_msd(
        frames, group_indices, lagmax, start, end, 
        timeintv: float, prefix=""
    ):
    """Compute MSD ...

    Args:
        group_func: Function used to get a group of atoms.

    """
    # -
    frames = wrap_traj(frames)
    frames = frames[start:end:]
    #print("nframes: ", len(frames))

    positions = []
    for atoms in frames:
        positions.append(atoms.get_positions()[group_indices,:])
    positions = np.array(positions)

    nframes, natoms, _ = positions.shape
    #print("shape: ", positions.shape)

    # -
    msds_by_particle = np.zeros((lagmax, natoms))

    lagtimes = np.arange(1, lagmax)
    for lag in lagtimes:
        disp = positions[:-lag, :, :] - positions[lag:, :, :]
        #print("disp: ", disp.shape)
        sqdist = np.square(disp).sum(axis=-1)
        msds_by_particle[lag, :] = np.mean(sqdist, axis=0)
    timeseries = msds_by_particle.mean(axis=1)

    timeintv = timeintv / 1000. # fs to ps
    lagtimes = np.arange(lagmax)*timeintv

    return lagtimes, timeseries


class DiffusionCoefficientValidator(AbstractValidator):

    """Estimate the diffusion coefficient.

    Compute a windowed MSD where the MSD is averaged over all possible lag-times
    tau < tau_max.

    """

    def __init__(
            self, group, timeintv: float, lagmax: int, start: int=None, end: int=None, 
            d_start: int = 0, d_end: int = 20,
            directory: Union[str, pathlib.Path] = "./", *args, **kwargs
        ):
        """"""
        super().__init__(directory, *args, **kwargs)

        self.start = start
        self.end = end
        
        # - diffusion coefficient linear fitting
        self.d_start = d_start
        self.d_end = d_end

        self.lagmax = lagmax
        self.timeintv = timeintv

        self.group = group

        return

    def _process_data(self, data) -> List[List[Atoms]]:
        """"""
        data = AtomsNDArray(data)
        self._debug(f"data: {data}")

        if data.ndim == 1:
            data = [data.tolist()]
        elif data.ndim == 2: # assume it is from minimisations...
            data = data.tolist()
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data
    
    def run(self, dataset: dict, worker=None, *args, **kwargs):
        """"""
        super().run()

        # - find some optional parameters
        labels = kwargs.get("labels", None)

        # - 
        self._print("process reference ->")
        reference = dataset.get("reference")
        if reference is not None:
            self._irun(reference, "ref-", labels)

        self._print("process prediction ->")
        prediction = dataset.get("prediction")
        if prediction is not None:
            self._irun(prediction, "pre-", labels)

        return
    
    def _irun(self, data, prefix="", labels=None):
        """Test the first trajectory.

        TODO: Several trajectories.

        """
        mdtrajs = self._process_data(data)
        group_indices = create_a_group(mdtrajs[0][0], self.group)
        self._debug(f"group_indices: {group_indices}")

        cache_msd = self.directory/f"{prefix}msd.npy"
        if not cache_msd.exists():
            data = Parallel(n_jobs=self.njobs)(
                delayed(compute_msd)(
                    [a for a in frames if a is not None], # AtomsNDArray may have None...
                    group_indices, lagmax=self.lagmax, 
                    start=self.start, end=self.end, timeintv=self.timeintv, 
                    prefix=prefix
                ) for frames in mdtrajs
            )
            np.save(cache_msd, data)
        else:
            data = np.load(cache_msd)

        lagtimes = [x[0] for x in data]
        timeseries = [x[1] for x in data]
        plot_msd(
            self.directory, names=labels, lagtimes=lagtimes, timeseries=timeseries, 
            start_step=self.d_start, end_step=self.d_end, prefix=prefix
        )

        return
    



if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*


import copy
import itertools
import pathlib
import shutil
from typing import Union, List, Optional


from ase.io import read, write


from . import AbstractPotentialManager, AbstractTrainer
from . import DummyCalculator, CommitteeCalculator


class MaceDataloader:

    name: str = "mace"

    def __init__(
        self,
        train_file: Union[str, pathlib.Path],
        test_file: Union[str, pathlib.Path],
        batchsize: int,
        directory: Union[str, pathlib.Path] = "./",
        *args,
        **kwargs,
    ) -> None:
        """"""
        self.train_file = pathlib.Path(train_file).resolve()
        self.test_file = pathlib.Path(test_file).resolve()

        self.batchsize = batchsize
        self.directory = pathlib.Path(directory).resolve()

        return

    def as_dict(
        self,
    ) -> dict:
        """"""
        params = {}
        params["name"] = self.name
        params["train_file"] = str(self.train_file)
        params["test_file"] = str(self.test_file)
        params["batchsize"] = self.batchsize
        params["directory"] = str(self.directory.resolve())

        return params


class MaceTrainer(AbstractTrainer):

    name = "mace"
    command = "mace_run_train"
    freeze_command = ""

    _train_fname = "_train.xyz"
    _test_fname = "_test.xyz"

    #: Flag indicates that the training is finished properly.
    CONVERGENCE_FLAG: str = "Done"

    def __init__(
        self,
        config: dict,
        type_list: List[str] = None,
        train_epochs: int = 200,
        print_epochs: int = 5,
        directory=".",
        command="python ./run_train.py",
        freeze_command="python ./run_train.py",
        random_seed: int = None,
        *args,
        **kwargs,
    ) -> None:
        """"""
        super().__init__(
            config,
            type_list,
            train_epochs,
            print_epochs,
            directory,
            command,
            freeze_command,
            random_seed,
            *args,
            **kwargs,
        )

        self._type_list = type_list

        return

    def _resolve_train_command(self, init_model=None, *args, **kwargs) -> str:
        """"""

        return self.command

    def freeze(self):
        """"""
        # models = list((self.directory/"checkpoints").glob("*.model"))
        use_swa = self.config.get("swa", False)
        if not use_swa:
            model_fpath = self.directory / ("{}.model".format(self.config["name"]))
        else:
            model_fpath = self.directory / ("{}_swa.model".format(self.config["name"]))

        return model_fpath

    def _resolve_freeze_command(self, *args, **kwargs) -> str:
        """"""

        return self.freeze_command

    @property
    def frozen_name(self):
        """"""

        return f"{self.name}.model"

    def _update_config(self, dataset, *args, **kwargs) -> dict:
        """"""
        # - update config
        train_config = copy.deepcopy(self.config)
        train_config["name"] = train_config.get("name", "mace")
        train_config["seed"] = self.random_seed
        train_config["train_file"] = str(dataset.train_file)
        train_config["valid_file"] = str(dataset.test_file)
        train_config["valid_fraction"] = 0.0
        test_file = train_config.get("test_file")
        if test_file is not None:
            train_config["test_file"] = str(pathlib.Path(test_file).resolve())
        else:
            train_config["test_file"] = str(dataset.test_file)

        # TODO: plus one to save the final checkpoint?
        train_config["max_num_epochs"] = self.train_epochs
        train_config["eval_interval"] = self.print_epochs
        train_config["batch_size"] = dataset.batchsize

        swa = train_config.get("swa", False)
        if swa:
            start_swa = train_config.get("start_swa", -1)
            if not (0 < start_swa < self.train_epochs):
                raise RuntimeError(
                    f"{start_swa = } must be smaller than {self.train_epochs = }"
                )
        else:
            ...

        # - misc
        import torch

        train_config["device"] = "cuda" if torch.cuda.is_available() else "cpu"

        restart_latest = train_config.get("restart_latest", True)
        train_config["restart_latest"] = restart_latest

        init_latest = train_config.get("init_latest", None)
        if init_latest is not None and init_latest:
            train_config["init_latest"] = init_latest
            del train_config["restart_latest"]

        train_config["save_cpu"] = True

        # - pop unused...
        train_config.pop("log_dir", None)
        train_config.pop("model_dir", None)
        train_config.pop("checkpoints_dir", None)
        train_config.pop("results_dir", None)

        return train_config

    def get_checkpoint(self):
        """"""

        return pathlib.Path(self.directory/"checkpoints").resolve()

    def _train_from_the_restart(self, dataset, init_model) -> str:
        """Train from the restart.

        Args:
            init_model: The path of the checkpoints folder.

        """
        # if init_model is not None:
        #     raise NotImplementedError(
        #         f"{self.name} does not support initialising from a previous model."
        #     )

        def _add_command_options(command, config) -> str:
            """"""
            # - convert to command line options...
            command = command + " "
            for k, v in config.items():
                if isinstance(v, bool):
                    if v:
                        command += f"--{k}  "
                elif isinstance(v, int) or isinstance(v, float):
                    command += f"--{k}={str(v)}  "
                else:
                    command += f"--{k}='{str(v)}'  "

            return command

        def _check_latest_checkpoint(ckpt_dir, model_name) -> Optional[int]:
            """"""
            ckpts = [p for p in ckpt_dir.glob(f"{model_name}*")]
            # ckpt_models = [c for c in ckpts if c.name.endswith(".model")]
            ckpt_models = [c for c in ckpts if c.name.endswith(".pt")]
            num_ckpts = len(ckpt_models)
            assert num_ckpts <= 1
            if num_ckpts == 1:
                ckpt_model = ckpt_models[0]
                prev_seed = int(ckpt_model.name.split("-")[1].split("_")[0])
            else:
                prev_seed = None

            return prev_seed

        def _get_latest_checkpoint(ckpt_dir, model_name):
            """"""
            ckpts = [p for p in ckpt_dir.glob(f"{model_name}*")]
            ckpt_models = [c for c in ckpts if c.name.endswith(".pt")]
            num_ckpts = len(ckpt_models)
            assert num_ckpts <= 1
            if num_ckpts == 1:
                ckpt_model = ckpt_models[0]
            else:
                ckpt_model = None

            return ckpt_model

        # Check dataset type and convert it if necessary
        dataset = self._prepare_dataset(dataset)

        train_config = self._update_config(dataset)

        # make command
        raw_command = self._train_from_the_scratch(dataset, init_model)

        ckpt_dir = self.directory / "checkpoints"
        if not self.directory.exists():
            if init_model is not None:
                model_name = train_config["name"]
                init_model = pathlib.Path(init_model)
                assert init_model.name == "checkpoints"
                self._print(f"init_model: {str(init_model)}")
                ckpt_path = _get_latest_checkpoint(init_model, model_name)
                if ckpt_path is not None:
                    ckpt_dir.mkdir()
                    curr_seed = train_config["seed"]
                    (ckpt_dir/f"{model_name}_run-{curr_seed}_epoch-0.pt").symlink_to(ckpt_path)
                    train_config.pop("restart_latest", None)
                    train_config["init_latest"] = True
                else:
                    self._print(f"FAILED to init from `{str(init_model)}`.")
            else:
                # train from the scratch and no config needs update
                ...
        else:
            if ckpt_dir.exists():
                # continue from the latest checkpoint
                prev_seed = _check_latest_checkpoint(ckpt_dir, train_config["name"])
                self._print(f"{prev_seed =}")
                if prev_seed is not None:
                    train_config["seed"] = prev_seed
                    train_config.pop("init_latest", None)
                    train_config["restart_latest"] = True
            else:
                if init_model is not None:
                    model_name = train_config["name"]
                    init_model = pathlib.Path(init_model)
                    assert init_model.name == "checkpoints"
                    self._print(f"init_model: {str(init_model)}")
                    ckpt_path = _get_latest_checkpoint(init_model, model_name)
                    if ckpt_path is not None:
                        ckpt_dir.mkdir()
                        curr_seed = train_config["seed"]
                        (ckpt_dir/f"{model_name}_run-{curr_seed}_epoch-0.pt").symlink_to(ckpt_path)
                        train_config.pop("restart_latest", None)
                        train_config["init_latest"] = True
                    else:
                        self._print(f"FAILED to init from `{str(init_model)}`.")
                else:
                    # train from the scratch and no config needs update
                    ...

        command = _add_command_options(raw_command, train_config)

        return command

    def _prepare_dataset(self, dataset, *args, **kwargs):
        """Prepare a reann dataset for training.

        Currently, it only supports converting xyz dataset.

        """
        self._print(f"{dataset = }")
        # NOTE: make sure the dataset path exists, sometimes it will be access
        #       before training to create a shared dataset
        self.directory.mkdir(parents=True, exist_ok=True)

        if not isinstance(dataset, MaceDataloader):
            set_names, train_frames, test_frames, adjusted_batchsizes = (
                dataset.split_train_and_test()
            )

            # NOTE: reann does not support split-system training,
            #       so we need merge all structures into one List
            train_frames = itertools.chain(*train_frames)
            write(self.directory / self._train_fname, train_frames)

            test_frames = itertools.chain(*test_frames)
            write(self.directory / self._test_fname, test_frames)

            dataset = MaceDataloader(
                train_file=self.directory / self._train_fname,
                test_file=self.directory / self._test_fname,
                directory=self.directory,
                batchsize=dataset.batchsize,
            )
        else:
            ...

        return dataset

    def write_input(self, dataset, *args, **kwargs):
        """Convert dataset to the target format and write the configuration file if it has."""
        self._print(f"write {self.name} inputs...")

        # - convert dataset
        dataset = self._prepare_dataset(dataset)

        # - input config
        #   mace uses command line options

        return

    def read_convergence(self):
        """"""
        super().read_convergence()

        converged = False
        logs = list((self.directory / "logs").glob("*.log"))
        assert len(logs) == 1, "There should be only one log."
        with open(logs[0], "r") as fopen:
            lines = fopen.readlines()
        if self.CONVERGENCE_FLAG in lines[-1]:
            converged = True

        return converged


class MaceManager(AbstractPotentialManager):

    name = "mace"
    implemented_backends = ["ase", "jax"]

    valid_combinations = (
        ("ase", "ase"),
        ("jax", "ase"),
    )

    def __init__(self):
        """"""

        return

    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        # - parse params
        calc_params = copy.deepcopy(calc_params)

        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())
        type_list = calc_params.pop("type_list", [])

        type_map = {}
        for i, a in enumerate(type_list):
            type_map[a] = i

        # - model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = pathlib.Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))
        self.calc_params.update(model=models)

        precision = calc_params.pop("precision", "float32")

        estimate_uncertainty = calc_params.get("estimate_uncertainty", False)

        # - create specific calculator
        calc = DummyCalculator()
        if self.calc_backend == "ase":
            # return ase calculator
            try:
                import torch
                from mace.calculators import MACECalculator
            except:
                raise ModuleNotFoundError(
                    "Please install mace and torch to use the ase interface."
                )
            device = "cuda" if torch.cuda.is_available() else "cpu"
            # print(f"mace device: {device}")
            calcs = []
            for m in models:
                # print("device", torch.device("cuda" if torch.cuda.is_available() else torch.device("cpu")))
                curr_calc = MACECalculator(
                    model_paths=m,
                    device=device,
                    default_dtype=precision,
                )
                calcs.append(curr_calc)
            calc = self._process_committee(calcs, estimate_uncertainty)
        elif self.calc_backend == "jax":
            try:
                import jax
                from mace_jax.calculators.mace import MACEJAXCalculator
            except:
                raise ModuleNotFoundError(
                    "Please install mace-jax and jax to use the jax interface."
                )
            # calcs = []
            # for m in models:
            #     curr_calc = MACEJAXCalculator(
            #
            #     )
            raise NotImplementedError("The JAX backend for MACE is under development.")
        elif self.calc_backend == "lammps":
            raise NotImplementedError(
                "The LAMMPS backend for MACE is under development."
            )
        else:
            ...

        self.calc = calc

        return

    def _process_committee(self, calcs, estimate_uncertainty: bool):
        """"""
        if len(calcs) == 1:
            calc = calcs[0]
        elif len(calcs) > 1:
            if estimate_uncertainty:
                calc = CommitteeCalculator(calcs)
            else:
                calc = calcs[0]
        else: # Empty list
            calc = DummyCalculator()

        return calc

    def switch_uncertainty_estimation(self, status: bool = True):
        """Switch on/off the uncertainty estimation."""
        # NOTE: Sometimes the manager loads several models and supports uncertainty
        #       by committee but the user disables it. We need change the calc to
        #       the correct one as the loaded one is just a single calculator.
        if not hasattr(self, "calc"):
            raise RuntimeError(
                "Fail to switch uncertainty status as it does not have a calc."
            )
        # print(f"{self.calc}")

        # NOTE: make sure manager.as_dict() can have correct param
        self.calc_params["estimate_uncertainty"] = status

        # - convert calculator
        if self.calc_backend == "ase":
            if status:
                if isinstance(self.calc, CommitteeCalculator):
                    ...  # nothing to do
                else:  # reload models
                    self.register_calculator(self.calc_params)
            else:
                if isinstance(self.calc, CommitteeCalculator):
                    # TODO: save previous calc?
                    self.calc = self.calc.calcs[0]
                else:
                    ...
        elif self.calc_backend == "lammps":
            ...
        else:
            # TODO:
            # Other backends cannot have uncertainty estimation,
            # give a warning?
            ...

        return

    def remove_loaded_models(self, *args, **kwargs):
        """Loaded TF models should be removed before any copy.deepcopy operations."""
        self.calc.reset()
        if self.calc_backend == "ase":
            if isinstance(self.calc, CommitteeCalculator):
                for c in self.calc.calcs:
                    c.models = None
            else:
                self.calc.models = None
        else:
            ...

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
from typing import Union, List, Tuple, Mapping

import numpy as np

import ase
from ase import Atoms, units
from ase.build import molecule
from ase.collections import g2
from ase.io import read, write
from ase.neighborlist import NeighborList, natural_cutoffs
from ase.ga.utilities import closest_distances_generator

from ..core.operation import Operation
from ..data.array import AtomsNDArray

"""Some extra operations.
"""


def str2list_int(inp: str, convention: str = "lmp") -> List[int]:
    """Convert a string to a List of int.

    Args:
        inp: A string contains numbers and colons.
        convention: The input convention either `lmp` or `ase`.
                    lmp index starts from 1 and includes the last.

    Examples:
        >>> str2list_int("1:2 4:6", "lmp")
        >>> [0, 1, 3, 4, 5]
        >>> str2list_int("1:2 4:6", "ase")
        >>> [1, 4, 5]

    Returns:
        A List of integers.

    """
    ret = []
    for x in inp.strip().split():
        curr_range = list(map(int, x.split(":")))
        if len(curr_range) == 1:
            start, end = curr_range[0], curr_range[0]
        else:
            start, end = curr_range
        if convention == "lmp":
            ret.extend([i - 1 for i in list(range(start, end + 1))])
        elif convention == "ase":
            ret.extend(list(range(start, end)))
        else:
            ...

    # remove duplicates
    ret = sorted(list(set(ret)))

    return ret


def rotate_a_molecule(atoms, use_com: bool, rng):
    """"""
    atoms = copy.deepcopy(atoms)
    num_atoms = len(atoms)
    if not use_com:
        center = np.mean(atoms.positions, axis=0)
    else:
        center = atoms.get_center_of_mass()

    if num_atoms > 1:
        phi, theta, psi = 360 * rng.uniform(0, 1, 3)
        atoms.euler_rotate(phi=phi, theta=0.5 * theta, psi=psi, center=center)

    return atoms


def convert_string_to_atoms(species: str):
    """Convert a string to an Atoms object.

    Args:
        species: Species' name can be a chemical symbol or a chemical formula.

    """
    # - build adsorbate
    atoms = None
    if species in ase.data.chemical_symbols:
        atoms = Atoms(species, positions=[[0.0, 0.0, 0.0]])
    elif species in g2.names:
        atoms = molecule(species)
    else:
        raise ValueError(f"Fail to create species {species}")

    return atoms


def compute_molecule_number_from_density(molecular_mass, volume, density) -> int:
    """Compute the number of molecules in the region with a given density.

    Args:
        moleculer_mass: unit in g/mol.
        volume: unit in Ang^3.
        density: unit in g/cm^3.

    Returns:
        Number of molecules in the region.

    """
    number = (density / molecular_mass) * volume * units._Nav * 1e-24

    return int(number)


def check_overlap_neighbour(
    atoms: Atoms, covalent_ratio, custom_dmin_dict={}, excluded_pairs=[]
):
    """use neighbour list to check newly added atom is neither too close or too
    far from other atoms
    """
    atomic_numbers = atoms.get_atomic_numbers()
    cell = atoms.get_cell(complete=True)
    natoms = len(atoms)

    cov_min, cov_max = covalent_ratio
    dmin_dict = closest_distances_generator(set(atomic_numbers), cov_min)
    # print(f"{dmin_dict =}")
    # print(f"{custom_dmin_dict}")
    for k, v in custom_dmin_dict.items():
        dmin_dict[k] = v
    # print(f"{dmin_dict =}")
    nl = NeighborList(
        cov_max * np.array(natural_cutoffs(atoms)),
        skin=0.0,
        self_interaction=False,
        bothways=True,
    )
    nl.update(atoms)

    is_valid = True
    for i in range(natoms):
        nei_indices, nei_offsets = nl.get_neighbors(i)
        if len(nei_indices) > 0:
            for j, offset in zip(nei_indices, nei_offsets):
                distance = np.linalg.norm(
                    atoms.positions[i] - (atoms.positions[j] + np.dot(offset, cell))
                )
                if (i, j) not in excluded_pairs:
                    atomic_pair = (atomic_numbers[i], atomic_numbers[j])
                    if distance < dmin_dict[atomic_pair]:
                        is_valid = False
                        break
        else:
            # Find isolated atom which is not allowed...
            is_valid = False
            break

    return is_valid


def convert_composition_to_list(composition: dict, region) -> List[Tuple[Atoms, int]]:
    """"""
    # - define the composition of the atoms to optimise
    blocks = []
    for k, v in composition.items():
        k = convert_string_to_atoms(k)
        if isinstance(v, int):  # number
            v = v
        else:  # string command
            data = v.split()
            if data[0] == "density":
                v = compute_molecule_number_from_density(
                    np.sum(k.get_masses()), region.get_volume(), density=float(data[1])
                )
            else:
                raise RuntimeError(f"Unrecognised composition {k:v}.")
        blocks.append((k, v))

    # - check if there is any molecule
    for k, v in blocks:
        if len(k) > 1:
            use_tags = True
            break
    else:
        use_tags = False

    # atom_numbers = [] # atomic number of inserted atoms
    # for species, num in composition_blocks:
    #    numbers = []
    #    for s, n in ase.formula.Formula(species.get_chemical_formula()).count().items():
    #        numbers.extend([ase.data.atomic_numbers[s]]*n)
    #    atom_numbers.extend(numbers*num)

    return blocks


class remove_vacuum(Operation):

    cache: str = "cache_frames.xyz"

    def __init__(self, structures, thickness: float = 20.0, directory="./") -> None:
        """"""
        input_nodes = [structures]
        super().__init__(input_nodes, directory)

        self.thickness = thickness

        return

    def forward(self, structures) -> List[Atoms]:
        """Remove some vaccum of structures.

        Args:
            structures: List[Atoms] or AtomsNDArray.

        """
        super().forward()

        if isinstance(structures, AtomsNDArray):
            frames = structures.get_marked_structures()
        else:
            frames = structures

        # TODO: convert to atoms_array?
        cache_fpath = self.directory / self.cache
        if cache_fpath.exists():
            frames = read(cache_fpath, ":")
        else:
            for a in frames:
                a.cell[2, 2] -= self.thickness
            write(cache_fpath, frames)

        self.status = "finished"

        return frames


class reset_cell(Operation):

    cache: str = "cache_frames.xyz"

    def __init__(self, structures, cell, directory="./") -> None:
        """"""
        input_nodes = [structures]
        super().__init__(input_nodes, directory)

        self.cell = np.array(cell)

        return

    def forward(self, structures) -> List[Atoms]:
        """Remove some vaccum of structures.

        Args:
            structures: List[Atoms] or AtomsNDArray.

        """
        super().forward()

        if isinstance(structures, AtomsNDArray):
            frames = structures.get_marked_structures()
        else:
            frames = structures

        # TODO: convert to atoms_array?
        cache_fpath = self.directory / self.cache
        if cache_fpath.exists():
            frames = read(cache_fpath, ":")
        else:
            center_of_cell = np.sum(self.cell, axis=0) / 2.0
            for a in frames:
                com = a.get_center_of_mass()
                a.set_cell(self.cell)
                a.positions -= com - center_of_cell
            write(cache_fpath, frames)

        self.status = "finished"

        return frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import dataclasses
import pathlib

from typing import List

import numpy as np

from ase.calculators.calculator import Calculator
from ase.geometry import find_mic


def compute_center_of_mass(
    cell, masses, positions, saved_positions, scaled: bool = False, pbc: bool = True
):
    """Compute center_of_mass in the fractional space.

    The positions should be properly processed with find_mic.

    Args:
        positions: The cartesian coordinates of a group of atoms.

    """

    shift = positions - saved_positions
    curr_vectors, curr_distances = find_mic(shift, cell, pbc=True)

    shifted_positions = positions + curr_vectors

    # dcom/dx = masses/np.sum(masses)
    com = masses @ shifted_positions / np.sum(masses)

    if scaled:
        com = cell.scaled_positions(com)
        for i in range(3):  # FIXME: seleced pbc?
            com[i] %= 1.0
            com[i] %= 1.0  # need twice see ase test

    return com


def compute_com_energy_and_forces(
    cell, masses, com, saved_coms, sigma: float, omega: float
):
    """"""
    # - compute energy
    x, x_t = com, saved_coms
    x1 = x - x_t
    x2 = x1**2 / 2.0 / sigma**2  # uniform sigma?
    v = omega * np.exp(-np.sum(x2, axis=1))

    energy = v.sum(axis=0)

    # - compute forces
    # -- dE/dx
    dEdx = np.sum(-v[:, np.newaxis] * x1 / sigma**2, axis=0)[np.newaxis, :]

    m = (masses / np.sum(masses))[:, np.newaxis]

    c_inv = np.linalg.inv(cell)
    forces = -np.transpose(c_inv @ (m @ dEdx).T)

    return energy, forces


class CenterOfMassGaussianCalculator(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    def __init__(
        self,
        groups: List[List[int]],
        sigma: List[float] = [0.05, 0.05, 1e6],
        omega: float = 0.2,
        scaled: bool = True,
        pace: int = 1,
        **kwargs,
    ):
        """Init center of mass gaussian."""
        super().__init__(**kwargs)

        self.groups = groups

        self.sigma = np.array(sigma)
        self.omega = omega

        self.scaled = scaled

        self.pace = pace

        # - private
        self._num_steps = 0

        self._saved_positions = None

        self._history_records = [[] for _ in range(len(self.groups))]

        return

    @property
    def num_steps(self) -> int:
        """"""

        return self._num_steps

    def reset_metadata(self):
        """Reset some simulation-related private attributes."""
        self._num_steps = 0
        self._saved_positions = None
        self._history_records = [[] for _ in range(len(self.groups))]

        return

    def calculate(
        self,
        atoms=None,
        properties=["energy"],
        system_changes=["positions", "numbers", "cell"],
    ):
        """"""
        super().calculate(atoms, properties, system_changes)

        # - save the intial group atoms position
        #   treat them as a whole molecule
        #   positions should be processed before using this calculator
        if self._saved_positions is None:
            self._saved_positions = []
            positions = copy.deepcopy(atoms.get_positions())
            for g in self.groups:
                self._saved_positions.append([positions[i] for i in g])

            content = ("# {:>10s}" + "{:>12s}  "*4 + "{:<s} \n").format(
                "step", "bias", "com_0", "com_1", "com_2", "tags"
            )

            log_fpath = pathlib.Path(self.directory) / "info.log"
            with open(log_fpath, "w") as fopen:
                fopen.write(content)

        # print(f"{self.num_steps =}")
        # - compute center_of_mass
        cell = atoms.get_cell(complete=True)
        masses = atoms.get_masses()
        positions = atoms.get_positions()
        for i, g in enumerate(self.groups):
            com = compute_center_of_mass(
                atoms.cell,
                masses[g],
                positions[g],
                saved_positions=self._saved_positions[i],
                scaled=self.scaled,
                pbc=True,
            )
            if self.num_steps % self.pace == 0:
                self._history_records[i].append(com)

        # - compute energy and forces
        energy = 0.
        forces = np.zeros(atoms.positions.shape)

        step_energies = []
        for i, g in enumerate(self.groups):
            saved_coms = np.array(self._history_records[i])
            com = saved_coms[-1]
            curr_energy, curr_forces = compute_com_energy_and_forces(
                cell, masses[g], com, saved_coms, sigma=self.sigma, omega=self.omega
            )
            energy += curr_energy
            step_energies.append(energy)
            for k, f in zip(g, curr_forces):
                forces[k] += f
        self._write_step(step_energies)

        # -
        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces

        # - increase steps
        self._num_steps += 1

        return

    def _write_step(self, step_energies: List[float]):
        """"""
        content = ""
        for i, (g, ene, records) in enumerate(
            zip(self.groups, step_energies, self._history_records)
        ):
            content += ("{:>12d}  {:>12.4f}  " + "{:>12.4f} " * 3 + " {:<s}" + "\n").format(
                self._num_steps, ene, *records[-1], "_".join([str(x) for x in g])
            )

        log_fpath = pathlib.Path(self.directory) / "info.log"
        with open(log_fpath, "a") as fopen:
            fopen.write(content)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import pathlib
import argparse
from sys import prefix

import numpy as np

from ase.io import read, write
from ase.constraints import FixAtoms
from ase.calculators.singlepoint import SinglePointCalculator

from gdpx.selector.structure_selection import calc_feature, cur_selection, select_structures

"""
data route
Pt32 and Pt31 ->
= (2x2) surface = 
Pt111 -> FCC and HCP
= (3x3) surface = 
if still good ->
various oxides
"""

FULL_DATASET_PATH = "/users/40247882/scratch2/PtOx-dataset/"
# DESC_JSON_PATH = "/users/40247882/scratch2/oxides/input-jsons/soap_param.json"
DESC_JSON_PATH = "/mnt/scratch2/users/40247882/oxides/eann-main/Compressed/soap_param.json"

def read_frames(dataset_path, system_name, pattern):
    # read xyz dataset
    dataset_path = pathlib.Path(dataset_path)
    # system_name = "O0Pt32"
    system_path = dataset_path / system_name # 2126, 1920, muts be times of both 32 and 10 = 160
    print(str(system_path))

    sorted_paths = []
    for p in system_path.glob(pattern):
        sorted_paths.append(p)
    sorted_paths.sort()

    total_frames = []
    for p in sorted_paths:
        print(p)
        frames = read(p, ":")
        print("number of frames: ", len(frames))
        total_frames.extend(frames)
    print("Total number: ", len(total_frames))
    
    return total_frames

def calc_desc(frames):
    # calculate descriptor to select minimum dataset
    with open(DESC_JSON_PATH, "r") as fopen:
        desc_dict = json.load(fopen)
    desc_dict = desc_dict["soap"]

    # TODO: inputs
    njobs = 4

    cwd = pathlib.Path.cwd()
    features_path = cwd / "features.npy"
    if features_path.exists():
        print("use precalculated features...")
        features = np.load(features_path)
        assert features.shape[0] == len(frames)
    else:
        print('start calculating features...')
        features = calc_feature(frames, desc_dict, njobs, features_path)
        print('finished calculating features...')
    
    return features

def select_structures(
    system_name, features, num, zeta=-1, strategy="descent", index_map=None, prefix="",
    manually_selected = None
):
    """ 
    """
    # cur decomposition 
    cur_scores, selected = cur_selection(features, num, zeta, strategy)
    content = '# idx cur sel\n'
    for idx, cur_score in enumerate(cur_scores):
        stat = 'F'
        if idx in selected:
            stat = 'T'
        if index_map is not None:
            idx = index_map[idx]
        content += '{:>12d}  {:>12.8f}  {:>2s}\n'.format(idx, cur_score, stat) 
    with open(cwd / (prefix+"cur_scores.txt"), 'w') as writer:
        writer.write(content)

    # map selected indices
    if index_map is not None:
        selected = [index_map[s] for s in selected]
    if manually_selected is not None:
        selected.extend(manually_selected)
    np.save(cwd / (prefix+"indices.npy"), selected)

    selected_frames = []
    for idx, sidx in enumerate(selected):
        selected_frames.append(frames[int(sidx)])

    return selected_frames

def create_training():
    # train model ensemble and run MD calculation to test properties
    return

parser = argparse.ArgumentParser()
parser.add_argument(
    "-d", "--main_dir", default=FULL_DATASET_PATH,
    help = "system name"
)
parser.add_argument(
    "-n", "--name", 
    help = "system name"
)
parser.add_argument(
    "-p", "--pattern", default = "*.xyz",
    help = "xyz search pattern"
)
parser.add_argument(
    "-num", "--number", 
    default = 320, type=int,
    help = "number of selection"
)
parser.add_argument(
    "-c", "--count", required = True,
    type = int,
    help = "number of selection"
)
parser.add_argument(
    "-es", "--energy_shift", 
    default = 0.0, type = float,
    help = "add energy correction for each structure"
)
parser.add_argument(
    "--more", action="store_true",
    help = "select more structures from the rest"
)

args = parser.parse_args()

patword = args.pattern.strip("*.xyz")

cwd = pathlib.Path.cwd()
print("read frames from system directory")
frames = read_frames(args.main_dir, args.name, args.pattern)

MAX_AEAVG = -1.4
remove_unstable = True # if remove structures with large energy and forces
if remove_unstable:
    new_frames = []
    for idx, atoms in enumerate(frames):
        ae_avg = atoms.get_potential_energy() / len(atoms)
        if ae_avg < MAX_AEAVG:
            new_frames.append(atoms)
        else:
            print(
                "Find {} unstable structure with energy {}".format(idx, ae_avg)
            )
    frames = new_frames

# if use modulo number
#number = int((np.floor((len(frames)+10)/args.number)) * args.number)
#print("modulo number of frames: ", number)
#if number == 0:
#    print("no selected")
#    exit()
number = args.number

# write(patword+"-tot.xyz", frames)

prefix = "r" + str(args.count) + "-"
if args.count == 0:
    if number > 0:
        converged_frames = []
        converged_indices = []
        # add converged structures
        for idx, atoms in enumerate(frames):
            # add constraint to neglect forces of frozen atoms
            cons = FixAtoms(indices=[a.index for a in atoms if a.position[2] < 2.5])
            atoms.set_constraint(cons)
            max_force = np.max(np.fabs(atoms.get_forces()))
            if max_force < 0.05:
                converged_frames.append(atoms)
                converged_indices.append(idx)

        # select frames
        nconverged = len(converged_frames)
        print("Number of converged: ", nconverged)
        features = calc_desc(frames)
        selected_frames = select_structures(
            args.name, features, number-nconverged, prefix=prefix,
            manually_selected = converged_indices
        ) # include manually selected structures
        # selected_frames.extend(converged_frames)
        # TODO: sometims the converged one will be selected... so duplicate...

        if args.energy_shift > 0:
            es = args.energy_shift
            print(f"add {es} [eV] energy correction to each structure!!!")
            for atoms in selected_frames:
                new_energy = atoms.get_potential_energy() + es
                new_forces = atoms.get_forces(apply_constraint=False).copy()
                calc = SinglePointCalculator(atoms, energy=new_energy, forces=new_forces)
                atoms.calc = calc

        print("Writing structure file... ")
        write(cwd / (prefix+patword+'-sel.xyz'), selected_frames)
        print("")
else:
    merged_indices = []
    for i in range(args.count):
        previous_indices_path = pathlib.Path("r" + str(i) + "-indices.npy")
        indices = np.load(previous_indices_path).tolist()
        merged_indices.extend(indices)
    print(
        "Number of unique frames: ",
        len(set(merged_indices))
    )

    if previous_indices_path.exists() and args.more:
        features_path = cwd / "features.npy"
        features = np.load(features_path)

        assert len(frames) == features.shape[0]

        index_map = []
        rest_features = []
        rest_frames = []
        for idx, atoms in enumerate(frames):
            if idx not in merged_indices:
                rest_frames.append(atoms)
                rest_features.append(features[idx])
                index_map.append(idx)
        if len(rest_frames) < number:
            raise ValueError("Not enough structures left...")
        rest_features = np.array(rest_features)
        selected_frames = select_structures(args.name, rest_features, number, index_map=index_map, prefix=prefix)

        print("Writing structure file... ")
        write(cwd / (prefix+patword+'-sel.xyz'), selected_frames)
        print("")


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import pathlib
from typing import Union, Callable

from .. import config


class Operation(abc.ABC):
    """"""

    identifier: str = "op"

    #: Working directory for the operation.
    _directory: Union[str, pathlib.Path] = pathlib.Path.cwd()

    #: Whether re-compute this operation
    status: str = "unfinished"  # ["unfinished", "ready", "wait", "finished"]

    #: Standard print function.
    _print: Callable = config._print

    #: Standard debug function.
    _debug: Callable = config._debug

    def __init__(self, input_nodes=[], directory: Union[str, pathlib.Path] = "./") -> None:
        """"""
        if hasattr(self, "_preprocess_input_nodes"):
            self.input_nodes = self._preprocess_input_nodes(input_nodes)
        else:
            self.input_nodes = input_nodes

        self.directory = directory

        # Initialize list of consumers
        # (i.e. nodes that receive this operation's output as input)
        self.consumers = []

        # Append this operation to the list of consumers of all input nodes
        for input_node in self.input_nodes:
            input_node.consumers.append(self)

        return

    @property
    def directory(self) -> pathlib.Path:
        """"""

        return self._directory

    @directory.setter
    def directory(self, directory_) -> None:
        """"""
        self._directory = pathlib.Path(directory_)

        return

    def reset(self):
        """Reset node's output and status."""
        if hasattr(self, "output"):
            delattr(self, "output")
            self.status = "unfinished"

        return
    
    def is_about_to_exit(self) -> bool:
        """Check whether this operation has an input is about to exit."""
        status = [node.status == "exit" for node in self.input_nodes]
        if any(status):
            return True
        else:
            return False

    def is_ready_to_forward(self) -> bool:
        """Check whether this operation is ready to forward."""
        # - check input nodes' status
        status = [node.status == "finished" for node in self.input_nodes]
        if all(status):
            return True
        else:
            return False

    @abc.abstractmethod
    def forward(self):
        """"""
        # - set working directory and logger
        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np


import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
plt.style.use("presentation")


from ase.io import read, write


def project_positions(pos, positions):
    """"""
    sigma = 0.05
    intensity = np.sum(
        np.exp(-np.linalg.norm(pos - positions[:, :2], axis=1)**2/(2*sigma**2))
    )

    return intensity


def compute_nanoparticle_distribution():
    """"""
    atoms = read("/scratch/gpfs/jx1279/copper+alumina/dptrain/r9/_prod/Cu13+s001p32.xyz")

    # - find all copper atoms
    group_indices = [i for i, a in enumerate(atoms) if a.symbol == "Cu"]
    print(group_indices)

    # - 
    positions = atoms.positions[group_indices, :]
    print(positions)

    x, y = np.linspace(0, 15, 301), np.linspace(0, 15, 301)
    grid = np.meshgrid(x, y)
    pairs = np.array(list(zip(grid[0].flatten(), grid[1].flatten())))
    print(pairs)
    print(pairs.shape)

    intensities = [project_positions(pos, positions) for pos in pairs]
    print(intensities)

    # - plot
    fig, ax = plt.subplots(1, 1, figsize=(12, 8))
    cntr = ax.contour(x, y, np.reshape(intensities, (301, 301)))
    fig.colorbar(cntr)
    plt.savefig("./xxx.png")

    return


from jax import numpy as jnp
from jax import grad, value_and_grad

def circle_fit_kasa(positions):
    """"""
    n = positions.shape[0]

    #z = jnp.reshape(positions[:, 0]**2 + positions[:, 1]**2, (-1,1))
    z = positions[:, 0]**2 + positions[:, 1]**2
    print(f"z: {z}")

    xy1 = jnp.hstack((positions, jnp.ones((n, 1))))
    print(xy1)

    s = jnp.linalg.solve(xy1, z)

    #return (s[0]/2., s[1]/2., jnp.sqrt((s[0]**2+s[1]**2)/4.+s[2]))
    return jnp.sqrt((s[0]**2+s[1]**2)/4.+s[2])

dfn = value_and_grad(circle_fit_kasa, argnums=0)


def xxx():
    """"""
    positions = np.array(
        [[0, 0], [1, 1.1], [1, -0.9], [0, 2]], dtype=np.float32
        #[[0, 0], [1, 1.1], [0, 2]], dtype=np.float32
        #[[0, 0], [0, 2]], dtype=np.float32
    )
    
    params = circle_fit_kasa(positions=positions)
    print(params)

    #p = dfn(positions)
    #print(p)

    return


if __name__ == "__main__":
    #compute_nanoparticle_distribution()
    xxx()
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
from typing import Union, List, Mapping
from itertools import groupby

import numpy as np

from ase import Atoms
from ase.formula import Formula

from ..core.register import registers
from .constraints import convert_indices
#from gdpx.graph.creator import find_molecules

"""Utilities to create a group of atoms.

This module tries to mimic the behaviour of LAMMPS group command.

"""

class AbstractAtomicGroup(abc.ABC):

    def __init__(self) -> None:
        """"""
        ...
    
    @abc.abstractmethod
    def get_group_indices(self, atoms: Atoms) -> List[int]:
        """"""

        return


class SymbolGroup(AbstractAtomicGroup):

    def __init__(self, symbols: str) -> None:
        """"""
        super().__init__()

        self.symbols = symbols.strip().split()

        return
    
    def get_group_indices(self, atoms: Atoms) -> List[int]:
        """"""
        super().get_group_indices(atoms)

        group_indices = []
        for i, a in enumerate(atoms):
            if a.symbol in self.symbols:
                group_indices.append(i)

        return


def create_a_molecule_group(atoms: Atoms, group_command: str, use_tags=True) -> List[List[int]]:
    """Find molecules in the structure."""
    args = group_command.strip().split()
    #assert args[0] in ["tag", "molecule"], f"{args[0]} is not implemented."

    if args[0] in ["tags", "molecule"]:
        groups = []
        if args[0] == "tag":
            if "tags" in atoms.arrays:
                # --- find molecuels based on tags
                natoms = len(atoms)
                tags = atoms.get_tags() # copied already
                for k, g in groupby(range(natoms), key=lambda x: tags[x]):
                    atomic_indices = list(g)
                    #symbols = [atoms[i].symbol for i in atomic_indices]
                    #formula = Formula.from_list(symbols).format("hill")
                    if str(k) in args[1:]:
                        groups.append(atomic_indices)
            else:
                raise RuntimeError("Cant find tags in atoms.")
    
        #if args[0] == "molecule":
        #    target_molecule = args[1]
        #    molecules = [target_molecule]
        #    # --- find molecules with graph connectivity
        #    #raise RuntimeError("No tags in atoms.")
        #    valid_symbols = []
        #    for m in molecules:
        #        valid_symbols.extend(list(Formula(m).count().keys()))
        #    valid_symbols = set(valid_symbols)
        #    atomic_indices = create_a_group(atoms, "symbol {}".format(" ".join(valid_symbols)))
        #    fragments = find_molecules(atoms, atomic_indices)
        #    if target_molecule in fragments:
        #        groups = fragments[target_molecule]
        #    else:
        #        raise RuntimeError(f"Cant find molecule {target_molecule} in atoms.")
    else:
        # NOTE: use atomic group that equals one molecule
        groups = [create_a_group(atoms, group_command)]

    return groups


def create_a_group(atoms: Atoms, group_command: str) -> List[int]:
    """Create a group of atoms from a structure based on rules.

    Args:
        atoms: Input structure.
        group_command: Command that defines the group.

    Returns:
        List of atomic indices subject to the group command.
    
    """
    #print(group_command)
    if isinstance(group_command, str):
        args = group_command.strip().split()
        assert args[0] in ["index", "id", "region", "symbol", "tag"], f"{args[0]} is not implemented."
        nargs = len(args)
    else: # indices
        assert isinstance(group_command, list), "group command is not a list of int."
        args = ["index"]
        args.extend(group_command)

    group_indices = []

    # - (direct) index
    if args[0] == "index":
        group_indices = args[1:]

    # - id (atom index)
    if args[0] == "id":
        # NOTE: input file should follow lammps convention
        #       i.e. the index starts from 1
        group_indices = convert_indices(" ".join(args[1:]))

    # - region
    if args[0] == "region":
        region_cls = registers.get("region", args[1], convert_name=True)
        region = region_cls.from_str(" ".join(args[1:]))
        group_indices = region.get_contained_indices(atoms)

    # - symbol
    if args[0] == "symbol":
        selected_symbols = args[1:]
        for i, a in enumerate(atoms):
            if a.symbol in selected_symbols:
                group_indices.append(i)
    
    # - tag
    if args[0] == "tag":
        tag_indices = [int(i) for i in args[1:]]
        tags = atoms.get_tags()
        group_indices = [i for (i,t) in enumerate(tags) if t in tag_indices]

    return group_indices


def create_an_intersect_group(atoms, group_commands: List[str]) -> List[int]:
    """Create an intersect group of atoms based on commands.
    """
    # - init group from the first command
    group_indices = create_a_group(atoms, group_commands[0])

    # - intersect by other commands if have any
    for group_command in group_commands[1:]:
        cur_indices = create_a_group(atoms, group_command)
        # TODO: use data type set?
        temp_indices = [i for i in cur_indices if i in group_indices]
        group_indices = temp_indices
        ...

    return group_indices


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import abc
import copy
import itertools

from typing import NoReturn, List, Mapping

import numpy as np

from ase import Atoms
from ase import data

from . import registers


def get_tags_per_species(atoms: Atoms) -> Mapping[str,Mapping[int,List[int]]]:
    """Get tags per species.

    Example:

        .. code-block:: python

            >>> atoms = Atoms("PtPtPtCOCO")
            >>> tags = [0, 0, 0, 1, 1, 2, 2]
            >>> atoms.set_tags(tags)
            >>> get_tags_per_species(atoms)
            >>> {'Pt3': {0: [0,1,2]}, 'CO': {1: [3,4], 2: [5,6]}}

    """

    tags = atoms.get_tags() # default is all zero

    tags_dict = {} # species -> tag list
    for key, group in itertools.groupby(enumerate(tags), key=lambda x:x[1]):
        cur_indices = [x[0] for x in group]
        #print(key, " :", cur_indices)
        cur_atoms = atoms[cur_indices]
        formula = cur_atoms.get_chemical_formula()
        #print(formula)
        #print(key)
        if formula not in tags_dict:
            tags_dict[formula] = []
        tags_dict[formula].append([key, cur_indices])

    return tags_dict


class Region(abc.ABC):

    """The base class of region.

    Triclinic, sphere, cylinder.

    TODO: override __contains__?

    """

    def __init__(self, origin: List[float], *args, **kwargs):
        """"""
        self._origin = np.array(origin)

        return
    
    @abc.abstractmethod
    def from_str(command: str):
        """Init a region from the command"""

        return
    
    def get_contained_indices(self, atoms: Atoms):
        """"""
        indices_within_region = []
        for i, a in enumerate(atoms):
            if self._is_within_region(a.position):
                indices_within_region.append(i)

        return indices_within_region
    
    def get_random_positions(self, size=1, rng=np.random):
        """"""
        random_positions = []
        for i in range(size):
            ran_pos = self._get_a_random_position(rng)
            random_positions.append(ran_pos)
        random_positions = np.array(random_positions)

        return random_positions
    
    @abc.abstractmethod
    def _get_a_random_position(self, rng):
        """"""

        return
    
    @abc.abstractmethod
    def _is_within_region(self, position) -> bool:
        """Positions are normally atomic positions or molecular centre positions."""

        return
    
    def get_tags_dict(self, atoms: Atoms):
        """Get tags dict for atoms within the (entire) system"""

        return get_tags_per_species(atoms)
    
    def get_contained_tags_dict(self, atoms: Atoms, tags_dict: dict=None) -> Mapping[str,List[int]]:
        """"""
        # - find tags and compute cops
        if tags_dict is None:
            tags_dict_within_system = get_tags_per_species(atoms)
        else:
            tags_dict_within_system = tags_dict
        
        # - NOTE: get wrapped positions due to PBC
        positions = copy.deepcopy(atoms.get_positions(wrap=True))

        cops_dict = {}
        for key, tags_and_indices in tags_dict_within_system.items():
            for tag, curr_indices in tags_and_indices:
                cur_positions = positions[curr_indices]
                # TODO: Considering PBC, surface may have molecules across boundaries.
                cop = np.average(cur_positions, axis=0)
                if key not in cops_dict:
                    cops_dict[key] = []
                cops_dict[key].append([tag, cop])
        
        # - check 
        tags_dict_within_region = {}
        for key, tags_and_cops in cops_dict.items():
            #print(tags_and_cops)
            for tag, cop in tags_and_cops:
                if self._is_within_region(cop):
                    if key in tags_dict_within_region:
                        tags_dict_within_region[key].append(tag)
                    else:
                        tags_dict_within_region[key] = [tag]
 
        return tags_dict_within_region
    
    def get_empty_volume(self, atoms: Atoms, tags_dict: dict=None, ratio: float=1.0) -> float:
        """Empty volume = Region volume - total volume of atoms within region.

        This is not always correct since all atoms in the fragment are considered 
        within the region if their cop is in the region.
        
        """
        # - get atom indices with given tags
        if tags_dict is None:
            tags_dict = self.get_contained_tags_dict(atoms)
        tags_within_region = []
        for key, tags in tags_dict.items():
            tags_within_region.extend(tags)
        atomic_indices = [i for i, t in enumerate(atoms.get_tags()) if t in tags_within_region]

        # - get atoms' radii
        radii = np.array([data.covalent_radii[data.atomic_numbers[atoms[i].symbol]] for i in atomic_indices])
        radii *= ratio

        atoms_volume = np.sum([4./3.*np.pi*r**3 for r in radii])

        return self.get_volume() - atoms_volume
    
    @abc.abstractmethod
    def get_volume(self) -> float:
        """"""

        return
    
    #def __eq__(self, other):
    #    """"""
    #    return all(self.__dict__ == other.__dict__)

    @abc.abstractmethod
    def as_dict(self) -> dict:
        """"""

        return


class AutoRegion(Region):

    _curr_atoms: Atoms = None

    def __init__(self, origin: List[float]=[0.,0.,0.], atoms=None, *args, **kwargs) -> NoReturn:
        """"""
        super().__init__(origin, *args, **kwargs)

        self._curr_atoms = atoms

        return
    
    @staticmethod
    def from_str(command: str):
        """"""

        raise NotImplementedError("AutoRegion does support init_from_str.")
    
    def _get_a_random_position(self, rng=np.random):
        """"""
        if self._curr_atoms is None:
            raise RuntimeError(f"No atoms is attached to {self.__class__.__name__}")
        
        frac_pos = rng.uniform(0,1,3)
        ran_pos = np.dot(frac_pos, self._curr_atoms.get_cell())
        
        return ran_pos
    
    def _is_within_region(self, position) -> bool:
        """"""
        if self._curr_atoms is None:
            raise RuntimeError(f"No atoms is attached to {self.__class__.__name__}")

        is_in = False
        pos_ = position - self._origin
        frac_pos_ = np.dot(np.linalg.inv(self._curr_atoms.get_cell().T), pos_)
        if (
            0. <= np.modf(frac_pos_[0])[0] < 1. and
            0. <= np.modf(frac_pos_[1])[0] < 1. and
            0. <= np.modf(frac_pos_[2])[0] < 1.
        ):
            is_in = True

        return is_in
    
    def get_volume(self) -> float:
        """"""
        if self._curr_atoms is None:
            raise RuntimeError(f"No atoms is attached to {self.__class__.__name__}")

        return self._curr_atoms.get_volume()

    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "auto"

        return region_params
    

class CubeRegion(Region):

    def __init__(self, origin: List[float], boundary: List[float], *args, **kwargs):
        """"""
        super().__init__(origin=origin, *args, **kwargs)
        boundaries_ = np.array(boundary, dtype=np.float64)
        assert len(boundaries_) == 6, "Cubic region needs 6 numbers to define."
        self.boundaries = boundaries_

        return

    @staticmethod
    def from_str(command: str):
        """"""
        data = [float(x) for x in command.strip().split()[1:]]
        origin = data[:3]
        boundary = data[3:]

        return CubeRegion(origin, boundary)
    
    def _get_a_random_position(self, rng=np.random):
        """"""
        boundaries_ = copy.deepcopy(self.boundaries)
        boundaries_ = np.reshape(boundaries_, (2,3))

        ran_frac_pos = rng.uniform(0,1,3)
        ran_pos = (
            self._origin + boundaries_[0,:] +
            (boundaries_[0,:] - boundaries_[1,:])*ran_frac_pos
        )

        return ran_pos

    def _is_within_region(self, position) -> bool:
        """"""
        ox, oy, oz = self._origin
        (xl, yl, zl, xh, yh, zh) = self.boundaries

        position = np.array(position)

        is_in = False
        if (
            (ox+xl <= position[0] <= ox+xh) and
            (oy+yl <= position[1] <= oy+yh) and
            (oz+zl <= position[2] <= oz+zh)
        ):
            is_in = True

        return is_in

    def get_volume(self) -> float:
        """"""
        (xl, yl, zl, xh, yh, zh) = self.boundaries

        return (xh-xl)*(yh-yl)*(zh-zl)

    def __repr__(self) -> str:
        """"""
        content = f"{self.__class__.__name__} "
        content += f"origin {self.boundaries[:3]} "
        content += f"lmin {self.boundaries[3:6]} "
        content += f"lmax {self.boundaries[6:9]} "

        return content

    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "cube"
        region_params["origin"] = self._origin.tolist()
        region_params["boundary"] = self.boundaries.tolist()

        return region_params


class SphereRegion(Region):

    def __init__(self, origin: List[float], radius: float, *args, **kwargs):
        """"""
        super().__init__(origin=origin, *args, **kwargs)
        self._radius = radius

        return

    @staticmethod
    def from_str(command: str):
        """"""
        data = [float(x) for x in command.strip().split()[1:]]
        origin = data[:3]
        radius = data[3]

        return SphereRegion(origin, radius)
    
    def _get_a_random_position(self, rng):
        """"""
        ran_coord = rng.uniform(0,1,3)
        polar = np.array([self._radius, np.pi, np.pi]) * ran_coord
        r, theta, phi = polar

        ran_pos = np.array(
            [
                r*np.sin(theta)*np.cos(phi),
                r*np.sin(theta)*np.sin(phi),
                r*np.cos(theta)
            ]
        )
        ran_pos += self._origin

        return ran_pos
    
    def _is_within_region(self, position) -> bool:
        """"""
        is_in = False

        position = np.array(position)
        distance = np.linalg.norm(position-self._origin)
        if distance <= self._radius:
            is_in = True

        return is_in
    
    def get_volume(self):
        """"""

        return 4./3.*np.pi*self._radius**3

    def __repr__(self) -> str:
        """"""
        content = f"{self.__class__.__name__} "
        content += f"radius {self._radius} "
        content += f"volume {self.get_volume()} "

        return content

    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "sphere"
        region_params["origin"] = self._origin.tolist()
        region_params["radius"] = self._radius

        return region_params


class CylinderRegion(Region):

    """Region by a vertical cylinder.
    """

    def __init__(self, origin: List[float], radius: float, height: float, *args, **kwargs):
        """"""
        super().__init__(origin=origin, *args, **kwargs)

        self._radius = radius
        self._height = height

        return

    @staticmethod
    def from_str(command: str):
        """"""
        data = [float(x) for x in command.strip().split()[1:]]
        origin = data[:3]
        radius = data[3]
        height = data[4]

        return CylinderRegion(origin, radius, height)

    def _get_a_random_position(self, rng):
        """"""
        r, theta, h = rng.uniform(0,1,3) # r, theta, h

        ran_pos = np.array(
            [
                r*np.cos(theta),
                r*np.sin(theta),
                h
            ]
        )
        ran_pos += self._origin

        return ran_pos
    
    def _is_within_region(self, position) -> bool:
        """"""
        ox, oy, oz = self._origin

        is_in = False
        if oz <= position[2] <= oz+self._height:
            distance = np.linalg.norm(position[:2] - self._origin[:2])
            if distance <= self._radius:
                is_in = True

        return is_in
    
    def get_volume(self) -> float:
        """"""
        return np.pi*self._radius**2*self._height

    def __repr__(self) -> str:
        """"""
        content = f"{self.__class__.__name__} "
        content += f"radius {self._radius} "
        content += f"height   {self._height} "

        return content

    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "cylinder"
        region_params["origin"] = self._origin.tolist()
        region_params["radius"] = self._radius
        region_params["height"] = self._height

        return region_params


class LatticeRegion(Region):

    def __init__(self, origin: List[float], cell: List[float], *args, **kwargs):
        """"""
        super().__init__(origin=origin, *args, **kwargs)
        self._cell = np.reshape(cell, (3,3))

        return

    @staticmethod
    def from_str(command: str):
        """"""
        data = [float(x) for x in command.strip().split()[1:]]
        origin = data[:3]
        cell = data[3:]

        return LatticeRegion(origin, cell)

    def _get_a_random_position(self, rng):
        """"""
        ran_frac_coord = rng.uniform(0,1,3)
        ran_pos = np.dot(ran_frac_coord, self._cell)
        ran_pos += self._origin

        return ran_pos
    
    def _is_within_region(self, position) -> bool:
        """"""
        is_in = False
        pos_ = position - self._origin
        frac_pos_ = np.dot(np.linalg.inv(self._cell.T), pos_)
        if (
            0. <= np.modf(frac_pos_[0])[0] < 1. and
            0. <= np.modf(frac_pos_[1])[0] < 1. and
            0. <= np.modf(frac_pos_[2])[0] < 1.
        ):
            is_in = True

        return is_in
    
    def get_volume(self) -> float:
        """"""
        a, b, c = self._cell

        return np.dot(np.cross(a,b), c)
    
    def __repr__(self) -> str:
        """"""
        content = f"{self.__class__.__name__}\n"
        content += f"origin\n"
        content += ("  "+"{:<12.8f}  "*3+"\n").format(*self._origin)
        content += f"cell\n"
        content += (("  "+"{:<12.8f}  "*3+"\n")*3).format(*self._cell.flatten())

        return content

    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "lattice"
        region_params["origin"] = self._origin.tolist()
        region_params["cell"] = self._cell.tolist()

        return region_params


class SurfaceLatticeRegion(LatticeRegion):

    def __init__(self, origin: List[float], cell: List[float], *args, **kwargs):
        """"""
        super().__init__(origin, cell, *args, **kwargs)

        assert self._origin[0] == 0. and self._origin[1] == 0., "The x and y of origin should be ZERO."
        assert self._cell[2][0] == 0. and self._cell[2][1] == 0., "The x and y of cell 3rd vec should be ZERO."

        return

    @staticmethod
    def from_str(command: str):
        """"""
        data = [float(x) for x in command.strip().split()[1:]]
        origin = data[:3]
        cell = data[3:]

        return SurfaceLatticeRegion(origin, cell)

    def _is_within_region(self, position) -> bool:
        """"""
        is_in = False
        #vec1, vec2 = self._cell[0], self._cell[1]
        #normal = np.cross(vec1, vec2)
        #normal = normal/np.linalg.norm(normal)
        if self._origin[2] <= position[2] < self._origin[2] + self._cell[2][2]:
            is_in = super()._is_within_region(position)

        return is_in

    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "surface_lattice"
        region_params["origin"] = self._origin.tolist()
        region_params["cell"] = self._cell.tolist()

        return region_params
    

class SurfaceRegion(Region):

    def __init__(self, origin: List[float], normal: List[float], thickness: float, *args, **kwargs):
        """"""
        super().__init__(origin=origin, *args, **kwargs)
        self._normal = np.array(normal, dtype=float)
        self._thickness = thickness

        return

    @staticmethod
    def from_str(command: str):
        """"""
        data = [float(x) for x in command.strip().split()[1:]]
        origin = data[:3]
        normal = data[3:6]
        thickness = data[6:9]

        return SurfaceRegion(origin, normal, thickness)

    def _get_a_random_position(self, rng):
        """"""
        ran_frac_coord = rng.uniform(0,1,3)
        ran_pos = np.dot(ran_frac_coord, self._cell)
        ran_pos += self._origin

        raise NotImplementedError()
    
    def _is_within_region(self, position) -> bool:
        """"""
        is_in = False
        pos_ = position - self._origin
        frac_pos_ = np.dot(np.linalg.inv(self._cell.T), pos_)
        if (
            0. <= np.modf(frac_pos_[0])[0] < 1. and
            0. <= np.modf(frac_pos_[1])[0] < 1. and
            0. <= np.modf(frac_pos_[2])[0] < 1.
        ):
            is_in = True

        raise NotImplementedError()
    
    def get_volume(self) -> float:
        """"""
        a, b, c = self._cell

        #return np.dot(np.cross(a,b), c)
        raise NotImplementedError()
    
    def __repr__(self) -> str:
        """"""
        content = f"{self.__class__.__name__}\n"
        content += f"origin\n"
        content += ("  "+"{:<12.8f}  "*3+"\n").format(*self._origin)
        #content += f"cell\n"
        #content += (("  "+"{:<12.8f}  "*3+"\n")*3).format(*self._cell.flatten())

        return content


class IntersectRegion(Region):

    #: Maximum number of attempts to get a random position.
    MAX_ATTEMPTS: int = 1000

    def __init__(self, regions, origin=np.zeros(0), *args, **kwargs):
        """Initialise an IntersectRegion.

        Args:
            regions: A List of regions.
            origin: Region origin should be always zero.
        
        """
        super().__init__(origin, *args, **kwargs)

        # NOTE: Can sub-regions be intersect ones?
        self.regions = regions
        assert len(self.regions), "IntersectRegion supports only twp sub-regions."

        self._regions = []
        for r in copy.deepcopy(regions):
            shape = r.pop("method", None)
            curr_region = registers.create(
                "region", shape, convert_name=True, **r
            )
            self._regions.append(curr_region)

        return

    @staticmethod
    def from_str(command: str):

        raise NotImplementedError()
    
    def _get_a_random_position(self, rng):
        """"""
        for i in range(self.MAX_ATTEMPTS):
            r1, r2 = self._regions
            ran_pos = r1._get_a_random_position(rng)
            if not r2._is_within_region(ran_pos):
                break
        else:
            raise RuntimeError("Fail to get a random position in the IntersectRegion.")

        return ran_pos
    
    def _is_within_region(self, position) -> bool:
        """"""

        raise NotImplementedError()
    
    def get_volume(self) -> float:
        """"""

        raise NotImplementedError()
    
    def as_dict(self):
        """"""
        region_params = {}
        region_params["method"] = "intersect"
        region_params["origin"] = self._origin.tolist()
        region_params["regions"] = self.regions

        return region_params


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .bondpair import compute_distance_and_shift, get_bond_information


__all__ = ["compute_distance_and_shift", "get_bond_information"]


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import logging
import os
import re
import shutil
import pathlib
import pickle
import tarfile
from typing import NoReturn, List

import numpy as np

from ase import Atoms
from ase import data, units
from ase.io import read, write
from ase.formula import Formula
from ase.ga.utilities import closest_distances_generator

from .. import registers
from .. import SingleWorker, DriverBasedWorker
from .. import dict2str
from ..expedition import AbstractExpedition
from .operators import select_operator, parse_operators, save_operator, load_operator

"""This module tries to offer a base class for all MonteCarlo-like methods.
"""

MC_EARLYSTOP_FNAME = "MC_EARLY_STOPPED"


def convert_blmin_to_str(blmin: dict) -> str:
    """"""
    elements = []
    for k in blmin.keys():
        elements.extend(k)
    elements = set(elements)
    nelements = len(elements)

    index_map = {}
    for i, e in enumerate(elements):
        index_map[e] = i
    distance_map = np.zeros((nelements, nelements))
    for (i, j), dis in blmin.items():
        distance_map[index_map[i], index_map[j]] = dis

    symbols = [data.chemical_symbols[e] for e in elements]

    content = "Bond Distance Minimum\n"
    # content += "  covalent ratio: {}\n".format(covalent_min)
    content += "  " + " " * 4 + ("{:>6}  " * nelements).format(*symbols) + "\n"
    for i, s in enumerate(symbols):
        content += "  " + ("{:<4}" + "{:>8.4f}" * nelements + "\n").format(
            s, *list(distance_map[i])
        )

    return content


class MonteCarlo(AbstractExpedition):

    restart = False

    #: Prefix of the working directory.
    WDIR_PREFIX: str = "cand"

    #: Name of the MC trajectory.
    TRAJ_NAME: str = "mc.xyz"

    #: Name of the file stores MC information (operations).
    INFO_NAME: str = "opstat.txt"

    def __init__(
        self,
        builder: dict,
        operators: List[dict],
        convergence: dict,
        random_seed=None,
        dump_period: int = 1,
        ckpt_period: int = 100,
        restart: bool = False,
        directory="./",
        *args,
        **kwargs,
    ) -> None:
        """Parameters for Monte Carlo.

        Args:
            overwrite: Whether overwrite calculation directory.

        """
        super().__init__(
            directory=directory,
            random_seed=random_seed,
        )

        self.directory = directory
        self.dump_period = dump_period
        self.ckpt_period = ckpt_period
        self.restart = restart

        # - check system type
        if isinstance(builder, dict):
            builder_params = copy.deepcopy(builder)
            builder_method = builder_params.pop("method")
            builder = registers.create(
                "builder", builder_method, convert_name=False, **builder_params
            )
        else:
            builder = builder
        self.builder = builder

        # - create worker
        self.worker = None

        # - parse operators
        self.operators, self.op_probs = parse_operators(operators)

        # - parse convergence
        self.convergence = convergence
        if self.convergence.get("steps", None) is None:
            self.convergence["steps"] = 1

        return

    def _init_structure(self):
        """Initialise the input structure.

        Set proper tags and minimise the structure. Prepare `self.atoms`,
        `self.energy_stored`, and `self.start_step`.

        """
        step_wdir = self.directory / f"{self.WDIR_PREFIX}0"
        if not step_wdir.exists():
            # - prepare atoms
            self._print("===== MonteCarlo Structure =====\n")
            tags = self.atoms.arrays.get("tags", None)
            if tags is None:
                # default is setting tags by elements
                symbols = self.atoms.get_chemical_symbols()
                type_list = sorted(list(set(symbols)))
                new_tags = [
                    type_list.index(s) * 10000 + i for i, s in enumerate(symbols)
                ]
                self.atoms.set_tags(new_tags)
                self._print("set default tags by chemical symbols...")
            else:
                self._print("set attached tags from the structure...")

        # - run init
        self._print("===== MonteCarlo Initial Minimisation =====\n")
        # NOTE: atoms lost tags in optimisation
        #       TODO: move this part to driver?
        curr_tags = self.atoms.get_tags()

        self.atoms.info["confid"] = 0
        self.atoms.info["step"] = -1  # NOTE: remove step info

        # TODO: whether init driver?
        self.worker.wdir_name = step_wdir.name
        _ = self.worker.run([self.atoms])
        self.worker.inspect(resubmit=True)
        if self.worker.get_number_of_running_jobs() == 0:
            curr_frames = self.worker.retrieve()[0]
            # - update atoms
            curr_atoms = curr_frames[-1]
            self.energy_stored = curr_atoms.get_potential_energy()
            self._print(f"ene: {self.energy_stored}")
            self.atoms = curr_atoms
            self.atoms.set_tags(curr_tags)
            write(self.directory / self.TRAJ_NAME, self.atoms)

            # -
            self.start_step = 0

            # - log operator status
            with open(self.directory / self.INFO_NAME, "w") as fopen:
                fopen.write(
                    "{:<24s}  {:<24s}  {:<12s}  {:<12s}  {:<24s}  {:<24s}  \n".format(
                        "#Operator", "Info", "natoms", "Success", "prev_ene", "curr_ene"
                    )
                )
            step_converged = True
        else:
            step_converged = False

        return step_converged

    def _attach_bond_length_minimum_list(self):
        """Find possible elements in the simulation and build a bond-distance list."""
        type_list = []
        for op in self.operators:
            # TODO: wee need further unify the names here
            if hasattr(op, "particles"):
                for p in op.particles:
                    type_list.extend(list(Formula(p).count().keys()))
            elif hasattr(op, "reservoir"):
                type_list.extend(list(Formula(op.reservoir["species"]).count().keys()))
            else:
                ...
        type_list = list(set(type_list + self.atoms.get_chemical_symbols()))
        self._print(f"{type_list =}")
        unique_atomic_numbers = [data.atomic_numbers[a] for a in type_list]

        for op in self.operators:
            op.blmin = closest_distances_generator(
                atom_numbers=unique_atomic_numbers,
                ratio_of_covalent_radii=op.covalent_min,
            )

        return

    def run(self, *args, **kwargs):
        """Run MonteCarlo simulation."""
        # - some imported packages change `logging.basicConfig`
        #   and accidently add a StreamHandler to logging.root
        #   so remove it...
        for h in logging.root.handlers:
            if isinstance(h, logging.StreamHandler) and not isinstance(
                h, logging.FileHandler
            ):
                logging.root.removeHandler(h)

        # - check if it has a valid worker..
        assert self.worker is not None, "MC has not set its worker properly."
        if isinstance(self.worker, DriverBasedWorker):
            self._print("Convert a DriverBasedWorker to a SingleWorker.")
            self.worker = SingleWorker.from_a_worker(self.worker)
        assert isinstance(
            self.worker, SingleWorker
        ), f"{self.__class__.__name__} only supports SingleWorker (set use_single=True)."
        self.worker.directory = self.directory

        # - create an atoms during the run-time
        #   If it is created in init, it will be re-used in active-learning loop.
        #   Thus, we create a new one every run time.
        frames = self.builder.run()
        assert (
            len(frames) == 1
        ), f"{self.__class__.__name__} only accepts one structure."
        self.atoms = frames[0]

        # - prepare logger and output some basic info...
        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        # - show operator information
        self._print("===== MonteCarlo Operators (Modifiers) =====\n")

        # -- register bond list
        self._attach_bond_length_minimum_list()

        for op in self.operators:
            for x in str(op).split("\n"):
                self._print(x)
            for l in convert_blmin_to_str(op.blmin).split("\n"):
                self._print(l)
        self._print(f"normalised probabilities {self.op_probs}\n")

        # -- add print function to operators
        for op in self.operators:
            op._print = self._print
            op._debug = self._debug

        # NOTE: check if operators' regions are consistent
        #       though it works, unexpected results may occur
        # TODO: need rewrite eq function as compare array is difficult
        # noperators = len(self.operators)
        # for i in range(1,noperators):
        #    if self.operators[i].region != self.operators[i-1].region:
        #        raise RuntimeError(f"Inconsistent region found in op {i-1} and op {i}")

        # - start!!!
        converged = self.read_convergence()
        if not converged:
            # - init structure
            # --
            step_converged = False
            if not self._veri_checkpoint():
                step_converged = self._init_structure()
            else:
                step_converged = True
                self._load_checkpoint()

            if not step_converged:
                self._print("Wait structure to initialise.")
                return
            else:
                self.start_step += 1

            # - run mc steps
            curr_step = self.start_step  # start_step
            while True:
                # -- check exit-loop conditions
                if curr_step > self.convergence["steps"]:
                    self._print("Monte Carlo reaches the maximum step.")
                    break
                if (self.directory/MC_EARLYSTOP_FNAME).exists():
                    self._print("Monte Carlo reaches the earlystopping.")
                    break
                # -- run step
                step_state = self._irun(curr_step)
                # -- state-specific
                if step_state == "UNFINISHED":
                    self._print("Wait MC step to finish.")
                    break
                elif step_state == "FINISHED":
                    # -- save checkpoint
                    self._save_checkpoint(step=curr_step)
                    # -- clean up
                    if (
                        (self.directory / f"{self.WDIR_PREFIX}{curr_step}").exists()
                    ) and (curr_step % self.dump_period != 0):
                        shutil.rmtree(self.directory / f"{self.WDIR_PREFIX}{curr_step}")
                    curr_step += 1
                elif step_state == "FAILED":
                    self._print(f"RETRY STEP {curr_step}.")
                elif step_state == "EARLYSTOPPED":
                    # We need a file flag to indicate the simutlation is finshed
                    # when read_convergence is called.
                    with open(self.directory/MC_EARLYSTOP_FNAME, "w") as fopen:
                        fopen.write(f"{step_state =}")
                else:
                    ...  # This should not happen.
        else:
            self._print("Monte Carlo is converged.")

        return

    def _irun(self, i: int) -> str:
        """Run a single MC step.

        Each step has three status as FINISHED, UNFINISHED, and FAILED.

        """
        self._print(f"===== MC Step {i} =====")
        self._print(f"RANDOM_SEED:  {self.random_seed}")
        for l in dict2str(self.rng.bit_generator.state).split("\n"):
            self._print(l)

        step_wdir = self.directory / f"{self.WDIR_PREFIX}{i}"
        self.worker.wdir_name = step_wdir.name

        # - operate atoms
        curr_op = select_operator(self.operators, self.op_probs, self.rng)
        self._print(f"operator {curr_op.__class__.__name__}")
        curr_atoms = curr_op.run(self.atoms, self.rng)
        if curr_atoms:  # is not None
            # --- add info
            curr_atoms.info["confid"] = int(f"{i}")
            curr_atoms.info["step"] = -1  # NOTE: remove step info from driver
        else:
            self._print("FAILED to run operation...")

        # - run postprocess
        if curr_atoms is not None:
            # - TODO: save some info not stored by driver
            curr_tags = curr_atoms.get_tags()

            # - run postprocess (spc, min or md)
            _ = self.worker.run([curr_atoms], read_ckpt=True)
            self.worker.inspect(resubmit=True)
            if self.worker.get_number_of_running_jobs() == 0:
                curr_atoms = self.worker.retrieve()[0][-1]
                curr_atoms.set_tags(curr_tags)

                self.energy_operated = curr_atoms.get_potential_energy()
                self._print(f"post ene: {self.energy_operated}")

                # -- metropolis
                success = curr_op.metropolis(
                    self.energy_stored, self.energy_operated, self.rng
                )

                self._save_step_info(curr_op, success)

                # -- update atoms
                if success:
                    self.energy_stored = self.energy_operated
                    self.atoms = curr_atoms
                    self._print("success...")
                else:
                    self._print("failure...")
                write(self.directory / self.TRAJ_NAME, self.atoms, append=True)

                step_state = "FINISHED"

                # -- check earlystopping
                step_state = self._check_earlystop(self.atoms)
            else:
                step_state = "UNFINISHED"
        else:
            # save the previous structure as the current operation gives no structure.
            step_state = "FAILED"

        return step_state

    def _check_earlystop(self, atoms: Atoms) -> str:
        """Check whether earlystopping should be done to avoid unphysical structures.

        Returns:
            A state string `FINISHED` or `EARLYSTOPPED`.

        """
        es_dict = self.convergence.get("earlystop", None)
        if es_dict is not None:
            prop = es_dict.get("property", None)
            if prop == "energy_per_atom":
                ae_min, ae_max = es_dict["range"]
                energy_per_atom = atoms.get_potential_energy() / len(atoms)
                if ae_min <= energy_per_atom < ae_max:
                    es_state = "FINISHED"
                else:
                    es_state = "EARLYSTOPPED"
            else:
                raise NotImplementedError(f"Unknown earlystop: {es_dict =}.")
        else:
            es_state = "FINISHED"

        return es_state

    def _veri_checkpoint(self) -> bool:
        """Verify checkpoints."""
        ckpt_wdirs = list(self.directory.glob("checkpoint.*"))
        nwdirs = len(ckpt_wdirs)

        verified = True
        if nwdirs > 0:
            # TODO: check the directory is not empty
            ...
        else:
            verified = False

        return verified

    def _save_checkpoint(self, step):
        """Save the current Monte Carlo state."""
        if self.ckpt_period > 0 and (step % self.ckpt_period == 0):
            self._print("SAVE CHECKPOINT...")
            ckpt_wdir = self.directory / f"checkpoint.{step}"
            ckpt_wdir.mkdir(parents=True)
            # - save the structure
            write(ckpt_wdir / "structure.xyz", self.atoms)
            # - save operator state
            for i, op in enumerate(self.operators):
                save_operator(op, ckpt_wdir / f"op-{i}.ckpt")
            # - save the random state
            with open(ckpt_wdir / "rng.ckpt", "wb") as fopen:
                pickle.dump(self.rng.bit_generator.state, fopen)
        else:
            ...

        return

    def _load_checkpoint(self):
        """Load the current Monet Carlo state."""
        # - Find the latest checkpoint
        ckpt_wdir = sorted(
            self.directory.glob("checkpoint.*"),
            key=lambda x: int(str(x.name).split(".")[-1]),
        )[-1]
        step = int(ckpt_wdir.name.split(".")[-1])
        self._print(f"LOAD CHECKPOINT STEP {step}.")

        # -- load operators
        op_files = sorted(
            ckpt_wdir.glob("op-*.ckpt"),
            key=lambda x: int(str(x.name).split(".")[0][3:]),
        )

        saved_operators = []
        for i, op_file in enumerate(op_files):
            saved_operator = load_operator(op_file)
            saved_operators.append(saved_operator)
        self.operators = saved_operators

        self._attach_bond_length_minimum_list()

        # -- add print function to operators
        for op in self.operators:
            op._print = self._print
            op._debug = self._debug

        self._print("===== Saved MonteCarlo Operators (Modifiers) =====\n")
        for op in self.operators:
            for x in str(op).split("\n"):
                self._print(x)
        self._print(f"normalised probabilities {self.op_probs}\n")

        # -- load random state
        with open(ckpt_wdir / "rng.ckpt", "rb") as fopen:
            rng_state = pickle.load(fopen)
        self.rng.bit_generator.state = rng_state

        # -- load structure
        self.start_step = step
        self.atoms = read(ckpt_wdir / "structure.xyz")
        self.energy_stored = self.atoms.get_potential_energy()

        # -- reset mctraj
        mctraj = read(self.directory / self.TRAJ_NAME, f":{step+1}")
        write(self.directory / self.TRAJ_NAME, mctraj)

        return

    def _save_step_info(self, curr_op, success: bool):
        """"""
        extra_info = getattr(curr_op, "_extra_info", "-")
        with open(self.directory / self.INFO_NAME, "a") as fopen:
            fopen.write(
                "{:<24s}  {:<24s}  {:<12d}  {:<12s}  {:<24.4f}  {:<24.4f}  \n".format(
                    curr_op.__class__.__name__,
                    extra_info,
                    len(self.atoms),
                    str(success),
                    self.energy_stored,
                    self.energy_operated,
                )
            )

        return

    def read_convergence(self):
        """Check the convergence of MC.

        Currently, the only criteria is whether the simulation reaches the maximum
        steps.

        """
        converged = False
        if (self.directory / self.TRAJ_NAME).exists():
            mctraj = read(self.directory / self.TRAJ_NAME, ":")
            nframes = len(mctraj)
            # self.start_step = nframes
            if nframes > self.convergence["steps"]:
                converged = True
            if (self.directory/MC_EARLYSTOP_FNAME).exists():
                converged = True
        else:
            ...

        return converged

    def get_workers(self):
        """Get all workers used by this expedition."""
        if hasattr(self.worker.potter, "remove_loaded_models"):
            self.worker.potter.remove_loaded_models()

        # workers = []
        # for curr_wdir in wdirs:
        #    curr_worker = copy.deepcopy(self.worker)
        #    curr_worker.directory = curr_wdir.parent
        #    curr_worker.wdir_name = curr_wdir.name
        #    workers.append(curr_worker)
        curr_worker = copy.deepcopy(self.worker)
        curr_worker.directory = self.directory
        curr_worker._retrieve_mode = "all"

        return [curr_worker]

    def as_dict(self) -> dict:
        """"""
        engine_params = {}
        engine_params["method"] = "monte_carlo"
        engine_params["builder"] = self.builder.as_dict()
        engine_params["worker"] = self.worker.as_dict()
        engine_params["operators"] = []
        for op in self.operators:
            engine_params["operators"].append(op.as_dict())
        engine_params["dump_period"] = self.dump_period
        engine_params["ckpt_period"] = self.ckpt_period
        engine_params["convergence"] = self.convergence
        engine_params["random_seed"] = self.random_seed

        engine_params = copy.deepcopy(engine_params)

        return engine_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import os
import pathlib
from typing import NoReturn

from . import AbstractPotentialManager


class VaspManager(AbstractPotentialManager):

    name = "vasp"

    implemented_backends = ["vasp", "vasp_interactive"]
    valid_combinations = (
        # calculator, dynamics
        ("vasp", "vasp"), 
        ("vasp_interactive", "ase")
    )

    def __init__(self):

        return

    def _set_environs(self, pp_path, vdw_path) -> None:
        """Set files need for calculation.

        NOTE: pp_path and vdw_path may not exist since we would like to create a
              dummy calculator.

        """
        # ===== environs TODO: from inputs 
        # - ASE_VASP_COMMAND
        # pseudo 
        if "VASP_PP_PATH" in os.environ.keys():
            os.environ.pop("VASP_PP_PATH", "")
        os.environ["VASP_PP_PATH"] = pp_path

        # - vdw 
        vdw_envname = "ASE_VASP_VDW"
        if vdw_envname in os.environ.keys():
            _ = os.environ.pop(vdw_envname, "")
        os.environ[vdw_envname] = vdw_path

        return

    def register_calculator(self, calc_params):
        """"""
        # - check backends
        super().register_calculator(calc_params)

        # - some extra params
        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())

        # NOTE: whether check pp and vdw existence
        #       since sometimes we'd like a dummy calculator
        #       -- convert paths to absolute ones
        incar = calc_params.pop("incar", None)
        pp_path = calc_params.pop("pp_path", "")
        vdw_path = calc_params.pop("vdw_path", "")

        incar = str(pathlib.Path(incar).absolute())
        self.calc_params.update(incar=incar)
        pp_path = str(pathlib.Path(pp_path).absolute())
        self.calc_params.update(pp_path=pp_path)
        vdw_path = str(pathlib.Path(vdw_path).absolute())
        self.calc_params.update(vdw_path=vdw_path)

        if self.calc_backend == "vasp":
            # return ase calculator
            from ase.calculators.vasp import Vasp
            calc = Vasp(directory=directory, command=command)

            # - set some default electronic parameters
            calc.set_xc_params("PBE") # NOTE: since incar may not set GGA
            calc.set(lorbit=10)
            calc.set(gamma=True)
            calc.set(lreal="Auto")
            if incar is not None:
                calc.read_incar(incar)
            self._set_environs(pp_path, vdw_path)
            # - update residual params
            calc.set(**calc_params)
        elif self.calc_backend == "vasp_interactive":
            from vasp_interactive import VaspInteractive
            calc = VaspInteractive(directory=directory, command=command)
            # - set some default electronic parameters
            calc.set_xc_params("PBE") # NOTE: since incar may not set GGA
            calc.set(lorbit=10)
            calc.set(gamma=True)
            calc.set(lreal="Auto")
            if incar is not None:
                calc.read_incar(incar)
            # - set some vasp_interactive parameters
            calc.set(potim=0.0)
            calc.set(ibrion=-1)
            calc.set(ediffg=0)
            #calc.set(isif=3) # TODO: Does not support stress for now...
            self._set_environs(pp_path, vdw_path)
            # - update residual params
            calc.set(**calc_params)
        else:
            raise NotImplementedError(f"Unimplemented backend {self.calc_backend} for vasp.")
        
        self.calc = calc

        return


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from collections import namedtuple

from typing import NoReturn, Tuple, Union, Mapping, List

import numpy
import numpy as np

from ase import Atoms
from ase.io import read, write 

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    ...


def rms_dict(x_ref: List[float], x_pred: List[float]):
    """ Takes two datasets of the same shape 
        and returns a dictionary containing RMS error data
    """

    x_ref = np.array(x_ref)
    x_pred = np.array(x_pred)

    if np.shape(x_pred) != np.shape(x_ref):
        raise ValueError('WARNING: not matching shapes in rms')

    error_2 = (x_ref - x_pred) ** 2

    average = np.sqrt(np.average(error_2))
    std_ = np.sqrt(np.var(error_2))

    return {'rmse': average, 'std': std_}


def add_rmse_text(ax, x_rmse, x_name):
    """"""
    # add text about RMSE
    rmse_text = "RMSE:\n"
    for _rms, name in zip(x_rmse, x_name):
        rmse_text += "{:>6.3f}+-{:>6.3f} {:<4s}\n".format(_rms["rmse"], _rms["std"], name)

    ax.text(
        0.9, 0.1, rmse_text, transform=ax.transAxes, 
        fontsize=18, fontweight="bold", 
        horizontalalignment="right", verticalalignment="bottom"
    )

    return


def get_properties(frames: List[Atoms], other_props = [], apply_constraint=False):
    """Get properties of frames for comparison.

    Currently, only total energy and forces are considered.

    Returns:
        tot_symbols: shape (nframes,)
        tot_energies: shape (nframes,)
        tot_forces: shape (nframes,3)

    """
    tot_symbols, tot_energies, tot_forces = [], [], []

    for atoms in frames: # free energy per atom
        # -- basic info
        symbols = atoms.get_chemical_symbols()
        tot_symbols.extend(symbols)

        # -- energy
        energy = atoms.get_potential_energy() 
        tot_energies.append(energy)

        # -- force
        forces = atoms.get_forces(apply_constraint=apply_constraint)
        tot_forces.extend(forces.tolist())

    return tot_symbols, tot_energies, tot_forces


def plot_distribution(ax, x_ref, x_pred, x_name="data", x_types=None, weights=None):
    """"""
    x_ref, x_pred = np.array(x_ref), np.array(x_pred)
    assert x_ref.shape[0] == x_pred.shape[0], "Input data is inconsistent."

    if weights is None:
        weights = np.ones(x_ref.shape)
    weights = np.array(weights)
    assert x_ref.shape[0] == weights.shape[0], "Weight is inconsistent."

    # use flat shape otherwise hist will be separate
    x_diff = (x_pred - x_ref) / weights
    #print("x_diff: ", x_diff.shape)
    #if len(x_diff.shape) > 1:
    #    x_diff = np.linalg.norm(x_diff, axis=1)

    pmax, pmin = np.max(x_diff), np.min(x_diff)

    ax.set_ylabel("Probability Density")
    ax.set_xlabel("$\Delta$"+x_name)

    num_bins = 20
    if x_types is None:
        x_diff = x_diff.flatten()
        n, bins, patches = ax.hist(x_diff, num_bins, density=True, label=x_name)
    else:
        # -- per type
        x_types = np.array(x_types)
    
        types = sorted(set(x_types))
        for t in types:
            t_mask = np.array(x_types==t)
            x_diff_t = x_diff[t_mask]
            if len(x_diff.shape) > 1:
                x_diff_t = x_diff_t.flatten()
            n, bins, patches = ax.hist(x_diff_t, num_bins, density=True, label=t)

    ax.legend()

    return


def plot_parity(ax, x_ref, x_pred, x_name="data", x_types=None, weights=None):
    """Plots the distribution of energy per atom on the output vs the input."""
    # - convert data type
    x_ref, x_pred = np.array(x_ref), np.array(x_pred)
    assert x_ref.shape[0] == x_pred.shape[0], f"Input data is inconsistent {x_ref.shape} vs. {x_pred.shape}."

    if weights is None:
        weights = np.ones(x_ref.shape)
    weights = np.array(weights)
    assert x_ref.shape[0] == weights.shape[0], f"Weight is inconsistent {x_ref.shape} vs. {weights.shape}."

    x_ref /= weights
    x_pred /= weights

    # - get the appropriate limits for the plot
    pmax = np.max(np.array([x_ref,x_pred])) # property max
    pmin = np.min(np.array([x_ref,x_pred]))
    edge = (pmax-pmin)*0.05

    plim = (pmin - edge, pmax + edge)
    ax.set_xlim(plim)
    ax.set_ylim(plim)

    # add line of slope 1 for refrence
    ax.plot(plim, plim, c="k")

    # - set labels
    ax.set_title(x_name)

    ax.set_xlabel("Reference")
    ax.set_ylabel("Prediction")

    # - either plot points all togather or per type
    x_rmse = [rms_dict(x_ref, x_pred)]
    x_rmse_names = [x_name]

    if x_types is None:
        # -- all togather
        ax.scatter(x_ref, x_pred, label=x_name)
        # -- rmse text
        add_rmse_text(ax, x_rmse, x_rmse_names)
    else:
        # -- per type
        x_types = np.array(x_types)

        types = sorted(set(x_types))
        for t in types:
            t_mask = np.array(x_types==t)
            #print(t_mask)
            x_ref_t, x_pred_t = x_ref[t_mask], x_pred[t_mask]
            #print(t, x_ref_t.shape)
            ax.scatter(x_ref_t, x_pred_t, label=t)

            _rms = rms_dict(x_ref_t, x_pred_t)
            x_rmse.append(_rms)
            x_rmse_names.append(t)
        
        # -- rmse text
        add_rmse_text(ax, x_rmse, x_rmse_names)
    
    ax.legend()

    return x_rmse, x_rmse_names


def parity_plot(
        emulation: Tuple[str, Union[numpy.array,numpy.ndarray]], 
        reference: Tuple[str, Union[numpy.array,numpy.ndarray]], 
        ax, prop: Tuple[str, str] = ('energy','[eV/atom]'), title: str ='Parity Plot'
) -> NoReturn :
    """ Plots the distribution of energy per atom on the output vs the input"""
    # extract data
    xdata, ydata = emulation[1], reference[1]
    # scatter plot of the data
    ax.scatter(xdata, ydata)

    # get the appropriate limits for the plot
    pmax = np.max(np.array([xdata,ydata])) # property max
    pmin = np.min(np.array([xdata,ydata]))
    plim = (pmin - 0.05, pmax + 0.05)

    ax.set_xlim(plim)
    ax.set_ylim(plim)

    # add line of slope 1 for refrence
    ax.plot(plim, plim, c='k')

    # set labels
    ax.set_xlabel(prop[0]+' '+emulation[0]+' '+prop[1], fontsize=24)
    ax.set_ylabel(prop[0]+' '+reference[0]+' '+prop[1], fontsize=24)

    #set title
    ax.set_title(title, fontsize=24, fontweight='bold')

    # add text about RMSE
    _rms = rms_dict(xdata, ydata)
    rmse_text = 'RMSE:\n{:>.3f}+-{:>.3f} {:s}'.format(_rms['rmse'], _rms['std'], prop[1])
    ax.text(
        0.9, 0.1, rmse_text, transform=ax.transAxes, fontsize=24, fontweight='bold', 
        horizontalalignment='right', verticalalignment='bottom'
    )
    
    return

PropInfo = namedtuple(
    'PropInfo', ['xlabel', 'ylabel', 'title']
) 


def parity_plot_dict(
        emulation: Mapping[str, Union[numpy.array,numpy.ndarray]], 
        reference: Mapping[str, Union[numpy.array,numpy.ndarray]], 
        ax, prop_dict: dict = None, apex=[1.0], write_info=True
):
    """ Plots the distribution of energy per atom on the output vs the input"""
    if prop_dict is None:
        prop_dict = {}
        prop_dict['xlabel'] = 'x'
        prop_dict['ylabel'] = 'y'
        prop_dict['title'] = 'Parity Plot'
    
    rmse_results = {}

    rmse_text = ''
    names = emulation.keys()
    pmaxs, pmins = [], []
    for name in names:
        # extract data
        xdata, ydata = emulation[name], reference[name]
        # print(xdata)
        # print(ydata)
        # scatter plot of the data
        ax.scatter(xdata, ydata, label=name)
        # find max and min
        #print(np.max(xdata), np.max(ydata))
        pmaxs.append( np.max([np.max(xdata),np.max(ydata)]) )
        pmins.append( np.min([np.min(xdata),np.min(ydata)]) )
        # add text about RMSE
        _rms = rms_dict(xdata, ydata)
        rmse_results[name] = _rms
        rmse_text += 'RMSE:\n{:>.3f}+-{:>.3f} {:s}\n'.format(_rms['rmse'], _rms['std'], name)
    if write_info:
        ax.text(
            0.9, 0.1, rmse_text, transform=ax.transAxes, fontsize=24, fontweight='bold', 
            horizontalalignment='right', verticalalignment='bottom'
        )

    # get the appropriate limits for the plot
    pmax = np.max(pmaxs)
    pmin = np.min(pmins)
    plim = np.array((pmin - (pmax-pmin)*0.05, pmax + (pmax-pmin)*0.05))

    ax.set_xlim(plim)
    ax.set_ylim(plim)

    # add line of slope 1 for refrence
    for a in apex:
        if a == 1:
            ax.plot(plim, a*plim, c="k")
        else:
            ax.plot(plim, a*plim, ls="--", c="k", label = "${}\sigma$".format(a))

    # set labels
    ax.set_xlabel(prop_dict["xlabel"])
    ax.set_ylabel(prop_dict["ylabel"])

    #set title
    ax.set_title(prop_dict["title"])

    ax.legend()
    
    return rmse_results

def energy_plot(ener_in, ener_out, ax, title='Plot of energy'):
    """ Plots the distribution of energy per atom on the output vs the input"""
    # scatter plot of the data
    ax.scatter(ener_in, ener_out)

    # get the appropriate limits for the plot
    for_limits = np.array(ener_in +ener_out)
    elim = (for_limits.min() - 0.05, for_limits.max() + 0.05)
    ax.set_xlim(elim)
    ax.set_ylim(elim)

    # add line of slope 1 for refrence
    ax.plot(elim, elim, c='k')

    # set labels
    ax.set_ylabel('energy by GAP / eV')
    ax.set_xlabel('energy by VASP / eV')

    #set title
    ax.set_title(title)

    # add text about RMSE
    _rms = rms_dict(ener_in, ener_out)
    rmse_text = 'RMSE:\n' + str(np.round(_rms['rmse'], 3)) + ' +- ' + str(np.round(_rms['std'], 3)) + 'eV/atom'
    ax.text(0.9, 0.1, rmse_text, transform=ax.transAxes, fontsize='large', \
            horizontalalignment='right', verticalalignment='bottom')
    
    return

def force_plot(in_force, out_force, ax, symbol='H', title='Plot of force'):
    """ Plots the distribution of force components per atom 
        on the output vs the input only plots for the given atom type(s)"""
    # extract data for only one species
    in_force = in_force[symbol]
    out_force = out_force[symbol]

    # scatter plot of the data
    ax.scatter(in_force, out_force)

    # get the appropriate limits for the plot
    for_limits = np.array(in_force + out_force)
    flim = (for_limits.min() - 1, for_limits.max() + 1)
    ax.set_xlim(flim)
    ax.set_ylim(flim)

    # add line of
    ax.plot(flim, flim, c='k')

    # set labels
    ax.set_ylabel('force by GAP / (eV/Å)')
    ax.set_xlabel('force by VASP / (eV/Å)')

    #set title
    ax.set_title(title)

    # add text about RMSE
    _rms = rms_dict(in_force, out_force)
    rmse_text = 'RMSE:\n' + str(np.round(_rms['rmse'], 3)) + ' +- ' + str(np.round(_rms['std'], 3)) + 'eV/Å'
    ax.text(0.9, 0.1, rmse_text, transform=ax.transAxes, fontsize='large', horizontalalignment='right',
        verticalalignment='bottom')


def extract_energy_and_forces(atom_frames,calc=None,atomic=True):
    """
    Electronic free energy and Hellman-Feynman forces
    """
    energies_dft, forces_dft = [], {}
    energies_gap, forces_gap = [], {}

    for atoms in atom_frames: # free energy per atom
        # basic info
        symbols = atoms.get_chemical_symbols()
        if atomic:
            natoms = len(atoms)
        else:
            natoms = 1
        # energy
        free_energy = atoms.get_potential_energy(force_consistent=True) # electronic free energy
        energies_dft.append(free_energy/natoms)
        # force
        forces = atoms.get_forces()
        for sym, force in zip(symbols,forces):
            if sym in forces_dft.keys():
                forces_dft[sym].extend(force)
            else:
                forces_dft[sym] = list(force)
        if calc:
            # use quip to calculate gap predicted energy
            atoms.set_calculator(calc)
            free_energy = atoms.get_potential_energy(force_consistent=True) # electronic free energy
            energies_gap.append(free_energy/natoms)
            # force
            forces = atoms.get_forces()
            for sym, force in zip(symbols,forces):
                if sym in forces_gap.keys():
                    forces_gap[sym].extend(force)
                else:
                    forces_gap[sym] = list(force)

    return forces_dft, forces_gap, energies_dft, energies_gap


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import itertools
import pathlib

from typing import Optional, List, Tuple

import numpy as np

from ase import Atoms
from ase.data import atomic_numbers, covalent_radii
from ase.calculators.calculator import Calculator
from ase.neighborlist import NeighborList, natural_cutoffs


from ..utils import get_bond_information


def compute_bond_gaussian_energy_and_forces(
    vec_mic, dis, equ_dis, bstrain, saved_bstrains, sigma: float, omega: float
):
    """Compute bias energy and forces on a single bond strain."""
    # print(f"{bstrain =}")
    # print(f"{saved_bstrains =}")
    # - compute energy
    x, x_t = bstrain, saved_bstrains
    x1 = x - x_t
    x2 = x1**2 / 2.0 / sigma**2  # uniform sigma?
    v = omega * np.exp(-np.sum(x2, axis=1))

    energy = v.sum(axis=0)

    # - compute forces
    # -- dE/ds _ shape (1, cv_dim)
    dEds = np.sum(-v[:, np.newaxis] * x1 / sigma**2, axis=0)[np.newaxis, :]

    # -- ds/dc # cv gradient wrt coordinate
    dsdx = vec_mic / dis / equ_dis

    forces = np.sum(-dEds * dsdx, axis=0)

    return energy, forces


class BondGaussianCalculator(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    def __init__(
        self,
        bonds: List[str] = ["C", "H", "O"],
        cov_equi_ratio: float = 1.0,
        covalent_ratio: Tuple[float, float] = [0.8, 1.6],
        target_indices: Optional[List[int]] = None,
        sigma: float = 0.1,
        omega: float = -0.2,
        pace: int = 1,
        **kwargs,
    ):
        """"""
        super().__init__(**kwargs)

        self.sigma = sigma
        self.omega = omega

        self.pace = pace

        self.sequ = cov_equi_ratio
        self.smin, self.smax = covalent_ratio

        #: Equilibrium bond distance dict.
        if isinstance(bonds[0], str):
            symbols = bonds
            bonds = list(itertools.product(bonds, bonds))
        else:
            symbols_, bonds_ = [], []
            for i, j in bonds:
                symbols_.extend([i, j])
                bonds_.append((i, j))
                bonds_.append((j, i))
            symbols = list(set(symbols_))
            bonds = list(set(bonds_))
        self.symbols = symbols
        self.bonds = bonds

        radii = {s: cov_equi_ratio * covalent_radii[atomic_numbers[s]] for s in symbols}
        self.eqdis_dict = {k: radii[k[0]] + radii[k[1]] for k in bonds}

        #:
        if target_indices is not None:
            self.target_indices = target_indices
        else:
            self.target_indices = None

        # - private properties
        self.neighlist = None

        self._num_steps = 0

        self._initial_bonds = []

        self._reacted_bonds = []

        self._history_records = {}

        return

    @property
    def num_steps(self) -> int:
        """"""

        return self._num_steps

    def reset_metadata(self):
        """"""

        return

    def calculate(
        self,
        atoms=None,
        properties=["energy"],
        system_changes=["positions", "numbers", "cell"],
    ):
        """"""
        super().calculate(atoms, properties, system_changes)

        if self.num_steps == 0:
            # - write bias log
            content = ""
            content += f"# {self.sequ} [{self.smin} {self.smax}]\n"
            content += f"# {self.eqdis_dict}\n"
            content += f"# pace {self.pace} width {self.sigma} height {self.omega}\n"
            content += "# {:>10s}  {:>12s}  {:>12s}\n".format(
                "step", "num_biased", "num_reacted"
            )
            log_fpath = pathlib.Path(self.directory) / "info.log"
            with open(log_fpath, "w") as fopen:
                fopen.write(content)

            # - write event log
            content = "# reaction events\n"
            event_fpath = pathlib.Path(self.directory) / "event.log"
            with open(event_fpath, "w") as fopen:
                fopen.write(content)

        # - create a neighlist
        if self.neighlist is None:
            self.neighlist = NeighborList(
                self.smax * self.sequ * np.array(natural_cutoffs(atoms)),
                skin=0.0,
                self_interaction=False,
                bothways=False,
            )
        else:
            ...
        self.neighlist.update(atoms)

        # - find species
        if self.target_indices is None:
            self.target_indices = [
                i for i, a in enumerate(atoms) if a.symbol in self.symbols
            ]
            print(f"{self.target_indices =}")

        # - get initial bonds
        if self.num_steps == 0:
            bond_pairs, _, _, _ = get_bond_information(
                atoms,
                self.neighlist,
                self.eqdis_dict,
                covalent_min=0.8,
                target_indices=self.target_indices,
                # allowed_bonds=[("C", "O"), ("C", "H"), ("H", "O")],
                allowed_bonds=self.bonds
            )
            self._initial_bonds = bond_pairs
            print(f"{self._initial_bonds =}")

        energy, forces = self._compute_bias(atoms)

        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces

        self._write_step()

        self._num_steps += 1

        return

    def _compute_bias(self, atoms: Atoms):
        """"""
        # - find species
        bond_pairs, bond_distances, bond_shifts, equi_distances = get_bond_information(
            atoms,
            self.neighlist,
            self.eqdis_dict,
            covalent_min=self.smin,
            target_indices=self.target_indices,
            allowed_bonds=self.bonds,
        )
        # print(f"{bond_pairs =}")
        # print(f"{bond_distances =}")

        # - compute energy and forces
        num_bonds = len(bond_pairs)
        if num_bonds > 0:
            energy = 0.0
            forces = np.zeros((atoms.positions.shape))
            # -- add bond with the max strain to the reactive bond list
            bond_strains = (bond_distances - equi_distances) / equi_distances
            max_index = np.argmax(bond_strains)
            max_bond_pair = bond_pairs[max_index]
            if self.num_steps % self.pace == 0:
                if max_bond_pair not in self._initial_bonds:
                    if max_bond_pair in self._history_records:
                        self._history_records[max_bond_pair].append(
                            [bond_strains[max_index]]
                        )
                    else:
                        self._history_records[max_bond_pair] = [[bond_strains[max_index]]]
                    self._write_event(max_bond_pair, "biased")
            # print(f"{self._history_records =}")
            # -- apply gaussian bias on reactive bonds
            for i, bond_pair in enumerate(bond_pairs):
                if bond_pair in self._history_records:
                    b_i, b_j = bond_pair
                    bond_strain = bond_strains[i]
                    # check whether bond is formed and clear history if so
                    if bond_strain <= 0.05:
                        if bond_pair not in self._reacted_bonds:
                            self._reacted_bonds.append(bond_pair)
                            self._history_records[bond_pair] = []
                            self._write_event(bond_pair, "bond_formed")
                        continue
                    else:
                        # FIXME: thereacted bond becomes reactive again?
                        if bond_pair not in self._reacted_bonds:
                            ...
                        else:
                            if bond_strain > 0.2:
                                self._history_records[bond_pair].append([bond_strain])
                                self._write_event(bond_pair, "bond_broken")
                            else:
                                continue
                    vec_mic = atoms.positions[b_i] - (
                        atoms.positions[b_j] + bond_shifts[i]
                    )
                    saved_bstrains = np.array(self._history_records[bond_pair])
                    curr_energy, curr_forces = compute_bond_gaussian_energy_and_forces(
                        vec_mic,
                        bond_distances[i],
                        equi_distances[i],
                        bstrain=bond_strain,
                        saved_bstrains=saved_bstrains,
                        sigma=self.sigma,
                        omega=self.omega,
                    )
                    energy += curr_energy
                    forces[b_i] += curr_forces
                    forces[b_j] -= curr_forces
                    # print(f"{b_i} ~ {b_j}: {curr_forces =}")
        else:
            energy = 0.0
            forces = np.zeros((atoms.positions.shape))

        return energy, forces

    def _write_step(self):
        """"""
        num_biased = len(self._history_records)
        num_reacted = len(self._reacted_bonds)

        content = ""
        content += "{:>12d}  {:>12d}  {:>12d}\n".format(
            self.num_steps, num_biased, num_reacted
        )

        log_fpath = pathlib.Path(self.directory) / "info.log"
        with open(log_fpath, "a") as fopen:
            fopen.write(content)

        return
    
    def _write_event(self, bond_pair: Tuple[int, int], event: str):
        """"""
        pair_name = "_".join([str(x) for x in bond_pair])

        content = f"{self.num_steps:>8d}  {pair_name:>24s}  {event:>24s}\n"
        event_fpath = pathlib.Path(self.directory) / "event.log"
        with open(event_fpath, "a") as fopen:
            fopen.write(content)

        return


if __name__ == "__main__":
    ...
